import logging
from typing import Text, Dict, Any, List, Optional, TYPE_CHECKING, Tuple, cast

from rasa.shared.constants import DOCS_URL_NLU_BASED_SLOTS, IGNORED_INTENTS
import rasa.shared.utils.io
from rasa.shared.core.slots import ListSlot, Slot
from rasa.shared.nlu.constants import (
    ENTITIES,
    ENTITY_ATTRIBUTE_TYPE,
    ENTITY_ATTRIBUTE_ROLE,
    ENTITY_ATTRIBUTE_GROUP,
    ENTITY_ATTRIBUTE_VALUE,
    INTENT,
    NOT_INTENT,
    INTENT_NAME_KEY,
    TEXT,
)
from rasa.shared.core.constants import (
    ACTIVE_FLOW,
    ACTIVE_LOOP,
    REQUESTED_SLOT,
    SLOT_MAPPINGS,
    MAPPING_TYPE,
    SlotMappingType,
    MAPPING_CONDITIONS,
)

if TYPE_CHECKING:
    from rasa.shared.core.trackers import DialogueStateTracker
    from rasa.shared.core.domain import Domain
    from rasa.shared.nlu.training_data.message import Message
    from rasa.utils.endpoints import EndpointConfig


logger = logging.getLogger(__name__)


class SlotMapping:
    """Defines functionality for the available slot mappings."""

    @staticmethod
    def validate(mapping: Dict[Text, Any], slot_name: Text) -> None:
        """Validates a slot mapping.

        Args:
            mapping: The mapping which is validated.
            slot_name: The name of the slot which is mapped by this mapping.

        Raises:
            InvalidDomain: In case the slot mapping is not valid.
        """
        from rasa.shared.core.domain import InvalidDomain

        if not isinstance(mapping, dict):
            raise InvalidDomain(
                f"Please make sure that the slot mappings for slot '{slot_name}' in "
                f"your domain are valid dictionaries. Please see "
                f"{DOCS_URL_NLU_BASED_SLOTS} for more information."
            )

        try:
            mapping_type = SlotMappingType(mapping.get(MAPPING_TYPE))
        except ValueError:
            raise InvalidDomain(
                f"Your domain uses an invalid slot mapping of type "
                f"'{mapping.get(MAPPING_TYPE)}' for slot '{slot_name}'. Please see "
                f"{DOCS_URL_NLU_BASED_SLOTS} for more information."
            )

        validations: Dict[SlotMappingType, List[Text]] = {
            SlotMappingType.FROM_ENTITY: ["entity"],
            SlotMappingType.FROM_INTENT: ["value"],
            SlotMappingType.FROM_TRIGGER_INTENT: ["value"],
            SlotMappingType.FROM_TEXT: [],
            SlotMappingType.CUSTOM: [],
            SlotMappingType.FROM_LLM: [],
        }

        required_keys = validations[mapping_type]
        for required_key in required_keys:
            if mapping.get(required_key) is None:
                raise InvalidDomain(
                    f"You need to specify a value for the key "
                    f"'{required_key}' in the slot mapping of type '{mapping_type}' "
                    f"for slot '{slot_name}'. Please see "
                    f"{DOCS_URL_NLU_BASED_SLOTS} for more information."
                )

    @staticmethod
    def _get_active_loop_ignored_intents(
        mapping: Dict[Text, Any], domain: "Domain", active_loop_name: Text
    ) -> List[Text]:
        from rasa.shared.core.constants import ACTIVE_LOOP

        mapping_conditions = mapping.get(MAPPING_CONDITIONS)
        active_loop_match = True
        ignored_intents = []

        if mapping_conditions:
            match_list = [
                condition.get(ACTIVE_LOOP) == active_loop_name
                for condition in mapping_conditions
            ]
            active_loop_match = any(match_list)

        if active_loop_match:
            form_ignored_intents = domain.forms.get(active_loop_name, {}).get(
                IGNORED_INTENTS, []
            )
            ignored_intents = SlotMapping.to_list(form_ignored_intents)

        return ignored_intents

    @staticmethod
    def intent_is_desired(
        mapping: Dict[Text, Any],
        tracker: "DialogueStateTracker",
        domain: "Domain",
        message: Optional["Message"] = None,
    ) -> bool:
        """Checks whether user intent matches slot mapping intent specifications."""
        mapping_intents = SlotMapping.to_list(mapping.get(INTENT, []))
        mapping_not_intents = SlotMapping.to_list(mapping.get(NOT_INTENT, []))

        active_loop_name = tracker.active_loop_name
        if active_loop_name:
            mapping_not_intents = (
                mapping_not_intents
                + SlotMapping._get_active_loop_ignored_intents(
                    mapping, domain, active_loop_name
                )
            )

        if message is not None:
            intent = message.get(INTENT, {}).get("name")
        elif tracker.latest_message:
            intent = tracker.latest_message.intent.get(INTENT_NAME_KEY)
        else:
            intent = None

        intent_not_blocked = not mapping_intents and intent not in set(
            mapping_not_intents
        )

        return intent_not_blocked or intent in mapping_intents

    # helpers
    @staticmethod
    def to_list(x: Optional[Any]) -> List[Any]:
        """Convert object to a list if it isn't."""
        if x is None:
            x = []
        elif not isinstance(x, list):
            x = [x]

        return x

    @staticmethod
    def entity_is_desired(
        mapping: Dict[Text, Any],
        tracker: "DialogueStateTracker",
        message: Optional["Message"] = None,
    ) -> List[str]:
        """Checks whether slot should be filled by an entity in the input or not.

        Args:
            mapping: Slot mapping.
            tracker: The tracker.
            message: The message being processed.

        Returns:
            A list of matching values.
        """
        if message is not None:
            extracted_entities = message.get(ENTITIES, [])
            matching_values = [
                cast(Text, entity[ENTITY_ATTRIBUTE_VALUE])
                for entity in extracted_entities
                if entity.get(ENTITY_ATTRIBUTE_TYPE)
                == mapping.get(ENTITY_ATTRIBUTE_TYPE)
                and entity.get(ENTITY_ATTRIBUTE_GROUP)
                == mapping.get(ENTITY_ATTRIBUTE_GROUP)
                and entity.get(ENTITY_ATTRIBUTE_ROLE)
                == mapping.get(ENTITY_ATTRIBUTE_ROLE)
            ]
        elif tracker.latest_message and tracker.latest_message.text is not None:
            matching_values = list(
                tracker.get_latest_entity_values(
                    mapping.get(ENTITY_ATTRIBUTE_TYPE),
                    mapping.get(ENTITY_ATTRIBUTE_ROLE),
                    mapping.get(ENTITY_ATTRIBUTE_GROUP),
                )
            )
        else:
            matching_values = []

        return matching_values

    @staticmethod
    def check_mapping_validity(
        slot_name: Text,
        mapping_type: SlotMappingType,
        mapping: Dict[Text, Any],
        domain: "Domain",
    ) -> bool:
        """Checks the mapping for validity.

        Args:
            slot_name: The name of the slot to be validated.
            mapping_type: The type of the slot mapping.
            mapping: Slot mapping.
            domain: The domain to check against.

        Returns:
            True, if intent and entity specified in a mapping exist in domain.
        """
        if (
            mapping_type == SlotMappingType.FROM_ENTITY
            and mapping.get(ENTITY_ATTRIBUTE_TYPE) not in domain.entities
        ):
            rasa.shared.utils.io.raise_warning(
                f"Slot '{slot_name}' uses a 'from_entity' mapping "
                f"for a non-existent entity '{mapping.get(ENTITY_ATTRIBUTE_TYPE)}'. "
                f"Skipping slot extraction because of invalid mapping."
            )
            return False

        if (
            mapping_type == SlotMappingType.FROM_INTENT
            and mapping.get(INTENT) is not None
        ):
            intent_list = SlotMapping.to_list(mapping.get(INTENT))
            for intent in intent_list:
                if intent and intent not in domain.intents:
                    rasa.shared.utils.io.raise_warning(
                        f"Slot '{slot_name}' uses a 'from_intent' mapping for "
                        f"a non-existent intent '{mapping.get('intent')}'. "
                        f"Skipping slot extraction because of invalid mapping."
                    )
                    return False

        return True


def validate_slot_mappings(domain_slots: Dict[Text, Any]) -> None:
    """Raises InvalidDomain exception if slot mappings are invalid."""
    rasa.shared.utils.io.raise_warning(
        f"Slot auto-fill has been removed in 3.0 and replaced with a "
        f"new explicit mechanism to set slots. "
        f"Please refer to {DOCS_URL_NLU_BASED_SLOTS} to learn more.",
        UserWarning,
    )

    for slot_name, properties in domain_slots.items():
        mappings = properties.get(SLOT_MAPPINGS, [])

        for slot_mapping in mappings:
            SlotMapping.validate(slot_mapping, slot_name)


class SlotFillingManager:
    """Manages slot filling based on conversation context."""

    def __init__(
        self,
        domain: "Domain",
        tracker: "DialogueStateTracker",
        message: Optional["Message"] = None,
        action_endpoint: Optional["EndpointConfig"] = None,
    ) -> None:
        self.domain = domain
        self.tracker = tracker
        self.message = message
        self._action_endpoint = action_endpoint

    def is_slot_mapping_valid(
        self,
        slot_name: str,
        mapping_type: SlotMappingType,
        mapping: Dict[str, Any],
    ) -> bool:
        """Check if a slot mapping is valid."""
        return SlotMapping.check_mapping_validity(
            slot_name=slot_name,
            mapping_type=mapping_type,
            mapping=mapping,
            domain=self.domain,
        )

    def is_intent_desired(self, mapping: Dict[str, Any]) -> bool:
        """Check if the intent matches the one indicated in the slot mapping."""
        return SlotMapping.intent_is_desired(
            mapping=mapping,
            tracker=self.tracker,
            domain=self.domain,
            message=self.message,
        )

    def _verify_mapping_conditions(
        self, mapping: Dict[Text, Any], slot_name: Text
    ) -> bool:
        if mapping.get(MAPPING_CONDITIONS) and mapping[MAPPING_TYPE] != str(
            SlotMappingType.FROM_TRIGGER_INTENT
        ):
            if not self._matches_mapping_conditions(mapping, slot_name):
                return False

        return True

    def _matches_mapping_conditions(
        self, mapping: Dict[Text, Any], slot_name: Text
    ) -> bool:
        slot_mapping_conditions = mapping.get(MAPPING_CONDITIONS)

        if not slot_mapping_conditions:
            return True

        active_flow = self.tracker.active_flow

        if active_flow:
            return self._mapping_conditions_match_flow(
                active_flow, slot_mapping_conditions
            )

        # if we are not in a flow, we could be in a form
        return self._mapping_conditions_match_form(slot_name, slot_mapping_conditions)

    @staticmethod
    def _mapping_conditions_match_flow(
        active_flow: str,
        slot_mapping_conditions: List[Dict[str, str]],
    ) -> bool:
        active_flow_conditions = list(
            filter(lambda x: x.get(ACTIVE_FLOW) is not None, slot_mapping_conditions)
        )
        return any(
            [
                condition.get(ACTIVE_FLOW) == active_flow
                for condition in active_flow_conditions
            ]
        )

    def _mapping_conditions_match_form(
        self, slot_name: str, slot_mapping_conditions: List[Dict[str, str]]
    ) -> bool:
        if (
            self.tracker.is_active_loop_rejected
            and self.tracker.get_slot(REQUESTED_SLOT) == slot_name
        ):
            return False

        # check if found mapping conditions matches form
        for condition in slot_mapping_conditions:
            # we allow None as a valid value for active_loop
            # therefore we need to set a different default value
            active_loop = condition.get(ACTIVE_LOOP, "")

            if active_loop and active_loop == self.tracker.active_loop_name:
                condition_requested_slot = condition.get(REQUESTED_SLOT)
                if not condition_requested_slot:
                    return True
                if condition_requested_slot == self.tracker.get_slot(REQUESTED_SLOT):
                    return True

            if active_loop is None and self.tracker.active_loop_name is None:
                return True

        return False

    def _fails_unique_entity_mapping_check(
        self,
        slot_name: Text,
        mapping: Dict[Text, Any],
    ) -> bool:
        from rasa.core.actions.forms import FormAction

        if mapping[MAPPING_TYPE] != str(SlotMappingType.FROM_ENTITY):
            return False

        form_name = self.tracker.active_loop_name

        if not form_name:
            return False

        if self.tracker.get_slot(REQUESTED_SLOT) == slot_name:
            return False

        form = FormAction(form_name, self._action_endpoint)

        if slot_name not in form.required_slots(self.domain):
            return False

        if form.entity_mapping_is_unique(mapping, self.domain):
            return False

        return True

    def _is_trigger_intent_mapping_condition_met(
        self, mapping: Dict[Text, Any]
    ) -> bool:
        active_loops_in_mapping_conditions = [
            condition.get(ACTIVE_LOOP)
            for condition in mapping.get(MAPPING_CONDITIONS, [])
        ]

        trigger_mapping_condition_met = True

        if self.tracker.active_loop_name is None:
            trigger_mapping_condition_met = False
        elif (
            active_loops_in_mapping_conditions
            and self.tracker.active_loop_name is not None
            and (
                self.tracker.active_loop_name not in active_loops_in_mapping_conditions
            )
        ):
            trigger_mapping_condition_met = False

        return trigger_mapping_condition_met

    def extract_slot_value_from_predefined_mapping(
        self,
        mapping_type: SlotMappingType,
        mapping: Dict[Text, Any],
    ) -> List[Any]:
        """Extracts slot value if slot has an applicable predefined mapping."""
        if (
            self.message is None
            and self.tracker.has_bot_message_after_latest_user_message()
        ):
            # TODO: this needs further validation - not sure if this breaks something!!!

            # If the bot sent a message after the user sent a message, we can't
            # extract any slots from the user message. We assume that the user
            # message was already processed by the bot and the slot value was
            # already extracted (e.g. for a prior form slot).
            return []

        should_fill_entity_slot = mapping_type == SlotMappingType.FROM_ENTITY

        should_fill_intent_slot = mapping_type == SlotMappingType.FROM_INTENT

        should_fill_text_slot = mapping_type == SlotMappingType.FROM_TEXT

        trigger_mapping_condition_met = self._is_trigger_intent_mapping_condition_met(
            mapping
        )

        should_fill_trigger_slot = (
            mapping_type == SlotMappingType.FROM_TRIGGER_INTENT
            and trigger_mapping_condition_met
        )

        value: List[Any] = []

        if should_fill_entity_slot:
            value = SlotMapping.entity_is_desired(mapping, self.tracker, self.message)
        elif should_fill_intent_slot or should_fill_trigger_slot:
            value = [mapping.get("value")]
        elif should_fill_text_slot:
            value = [self.message.get(TEXT)] if self.message is not None else []
            if not value:
                value = [
                    self.tracker.latest_message.text
                    if self.tracker.latest_message is not None
                    else None
                ]

        return value

    def should_fill_slot(
        self, slot_name: str, mapping_type: SlotMappingType, mapping: Dict[Text, Any]
    ) -> bool:
        """Checks if a slot should be filled based on the conversation context."""
        if not self.is_slot_mapping_valid(slot_name, mapping_type, mapping):
            return False

        if not self.is_intent_desired(mapping):
            return False

        if not self._verify_mapping_conditions(mapping, slot_name):
            return False

        if self._fails_unique_entity_mapping_check(slot_name, mapping):
            return False

        return True


def extract_slot_value(
    slot: Slot, slot_filling_manager: SlotFillingManager
) -> Tuple[Any, bool]:
    """Extracts the value of a slot based on the conversation context."""
    is_extracted = False

    for mapping in slot.mappings:
        mapping_type = SlotMappingType(
            mapping.get(MAPPING_TYPE, SlotMappingType.FROM_LLM.value)
        )

        if mapping_type in [SlotMappingType.FROM_LLM, SlotMappingType.CUSTOM]:
            continue

        if not slot_filling_manager.should_fill_slot(slot.name, mapping_type, mapping):
            continue

        value: List[Any] = (
            slot_filling_manager.extract_slot_value_from_predefined_mapping(
                mapping_type, mapping
            )
        )

        if value:
            if not isinstance(slot, ListSlot):
                value = value[-1]

            if (
                value is not None
                or slot_filling_manager.tracker.get_slot(slot.name) is not None
            ):
                logger.debug(f"Extracted value '{value}' for slot '{slot.name}'.")

                is_extracted = True
                return value, is_extracted

    return None, is_extracted



import json
import logging
from typing import List, Text, Any, Dict, Optional

from rasa.core.constants import DEFAULT_REQUEST_TIMEOUT
from rasa.core.nlg.generator import NaturalLanguageGenerator, ResponseVariationFilter
from rasa.shared.core.trackers import DialogueStateTracker, EventVerbosity
from rasa.shared.exceptions import RasaException
from rasa.utils.endpoints import EndpointConfig

logger = logging.getLogger(__name__)


def nlg_response_format_spec() -> Dict[Text, Any]:
    """Expected response schema for an NLG endpoint.

    Used for validation of the response returned from the NLG endpoint.
    """
    return {
        "type": "object",
        "properties": {
            "text": {"type": "string"},
            "id": {"type": ["string", "null"]},
            "buttons": {"type": ["array", "null"], "items": {"type": "object"}},
            "elements": {"type": ["array", "null"], "items": {"type": "object"}},
            "attachment": {"type": ["object", "null"]},
            "image": {"type": ["string", "null"]},
            "custom": {"type": "object"},
        },
    }


RESPONSE_ID_KEY = "response_ids"


def nlg_request_format(
    utter_action: Text,
    tracker: DialogueStateTracker,
    output_channel: Text,
    **kwargs: Any,
) -> Dict[Text, Any]:
    """Create the json body for the NLG json body for the request."""
    tracker_state = tracker.current_state(EventVerbosity.ALL)
    response_id = kwargs.pop("response_id", None)

    return {
        "response": utter_action,
        "id": response_id,
        "arguments": kwargs,
        "tracker": tracker_state,
        "channel": {"name": output_channel},
    }


class CallbackNaturalLanguageGenerator(NaturalLanguageGenerator):
    """Generate bot utterances by using a remote endpoint for the generation.

    The generator will call the endpoint for each message it wants to
    generate. The endpoint needs to respond with a properly formatted
    json. The generator will use this message to create a response for
    the bot.
    """

    def __init__(self, endpoint_config: EndpointConfig) -> None:
        self.nlg_endpoint = endpoint_config

    async def generate(
        self,
        utter_action: Text,
        tracker: DialogueStateTracker,
        output_channel: Text,
        **kwargs: Any,
    ) -> Dict[Text, Any]:
        """Retrieve a named response from the domain using an endpoint."""
        domain_responses = kwargs.pop("domain_responses", None)
        response_id = self.fetch_response_id(
            utter_action, tracker, output_channel, domain_responses
        )
        kwargs["response_id"] = response_id

        body = nlg_request_format(utter_action, tracker, output_channel, **kwargs)

        logger.debug(
            "Requesting NLG for {} from {}. The request body is {}.".format(
                utter_action, self.nlg_endpoint.url, json.dumps(body)
            )
        )

        response = await self.nlg_endpoint.request(
            method="post", json=body, timeout=DEFAULT_REQUEST_TIMEOUT
        )

        logger.debug(f"Received NLG response: {json.dumps(response)}")

        if isinstance(response, dict) and self.validate_response(response):
            return response
        else:
            raise RasaException("NLG web endpoint returned an invalid response.")

    @staticmethod
    def validate_response(content: Optional[Dict[Text, Any]]) -> bool:
        """Validate the NLG response. Raises exception on failure."""
        from jsonschema import validate
        from jsonschema import ValidationError

        try:
            if content is None or content == "":
                # means the endpoint did not want to respond with anything
                return True
            else:
                validate(content, nlg_response_format_spec())
                return True
        except ValidationError as e:
            raise RasaException(
                f"{e.message}. Failed to validate NLG response from API, make sure "
                f"your response from the NLG endpoint is valid. "
                f"For more information about the format please consult the "
                f"`nlg_response_format_spec` function from this same module: "
                f"https://github.com/RasaHQ/rasa/blob/main/rasa/core/nlg/callback.py"
            )

    @staticmethod
    def fetch_response_id(
        utter_action: Text,
        tracker: DialogueStateTracker,
        output_channel: Text,
        domain_responses: Optional[Dict[Text, List[Dict[Text, Any]]]],
    ) -> Optional[Text]:
        """Fetch the response id for the utter action.

        The response id is retrieved from the domain responses for the
        utter action given the tracker state and channel.
        """
        if domain_responses is None:
            logger.debug("Failed to fetch response id. Responses not provided.")
            return None

        response_filter = ResponseVariationFilter(domain_responses)
        response_id = response_filter.get_response_variation_id(
            utter_action, tracker, output_channel
        )

        if response_id is None:
            logger.debug(f"Failed to fetch response id for action '{utter_action}'.")

        return response_id



from __future__ import annotations

import abc
import os
import shutil
from enum import Enum
from typing import TYPE_CHECKING, List, Optional, Text, Tuple, Union

import structlog

import rasa.shared.utils.common
import rasa.utils.common
from rasa.constants import (
    HTTP_STATUS_FORBIDDEN,
    HTTP_STATUS_NOT_FOUND,
    MODEL_ARCHIVE_EXTENSION,
)
from rasa.env import (
    AWS_ENDPOINT_URL_ENV,
    AZURE_ACCOUNT_KEY_ENV,
    AZURE_ACCOUNT_NAME_ENV,
    AZURE_CONTAINER_ENV,
    BUCKET_NAME_ENV,
    REMOTE_STORAGE_PATH_ENV,
)
from rasa.shared.exceptions import RasaException
from rasa.shared.utils.io import raise_warning

if TYPE_CHECKING:
    from azure.storage.blob import ContainerClient

structlogger = structlog.get_logger()


class RemoteStorageType(Enum):
    """Enum for the different remote storage types."""

    AWS = "aws"
    GCS = "gcs"
    AZURE = "azure"

    @classmethod
    def list(cls) -> List[str]:
        """Returns a list of all available storage types."""
        return [item.value for item in cls]


"""Storage can be a built-in one or a module path to a custom persistor."""
StorageType = Union[RemoteStorageType, str]


def parse_remote_storage(value: str) -> StorageType:
    try:
        return RemoteStorageType(value)
    except ValueError:
        # if the value is not a valid storage type,
        # but it is a string we assume it is a custom class
        # and return it as is

        supported_storages_help_text = (
            f"Supported storages are: {RemoteStorageType.list()} "
            "or path to a python class which implements `Persistor` interface."
        )

        if isinstance(value, str):
            if value == "":
                raise RasaException(
                    f"The value can't be an empty string."
                    f" {supported_storages_help_text}"
                )

            return value

        raise RasaException(
            f"Invalid storage type '{value}'. {supported_storages_help_text}"
        )


def get_persistor(storage: StorageType) -> Optional[Persistor]:
    """Returns an instance of the requested persistor.

    Currently, `aws`, `gcs`, `azure` and providing module paths are supported remote
    storages.
    """
    if storage == RemoteStorageType.AWS:
        return AWSPersistor(
            os.environ.get(BUCKET_NAME_ENV), os.environ.get(AWS_ENDPOINT_URL_ENV)
        )
    if storage == RemoteStorageType.GCS:
        return GCSPersistor(os.environ.get(BUCKET_NAME_ENV))

    if storage == RemoteStorageType.AZURE:
        return AzurePersistor(
            os.environ.get(AZURE_CONTAINER_ENV),
            os.environ.get(AZURE_ACCOUNT_NAME_ENV),
            os.environ.get(AZURE_ACCOUNT_KEY_ENV),
        )
    # If the persistor is not a built-in one, it is assumed to be a module path
    # to a persistor implementation supplied by the user.
    if storage:
        try:
            persistor = rasa.shared.utils.common.class_from_module_path(storage)
            return persistor()
        except ImportError:
            raise ImportError(
                f"Unknown model persistor {storage}. Please make sure to "
                f"either use an included model persistor ({RemoteStorageType.list()}) "
                f"or specify the module path to an external "
                "model persistor."
            )
    return None


class Persistor(abc.ABC):
    """Store models in cloud and fetch them when needed."""

    def persist(self, trained_model: str) -> None:
        """Uploads a trained model persisted in the `target_dir` to cloud storage."""
        file_key = self._create_file_key(trained_model)
        self._persist_tar(file_key, trained_model)

    def retrieve(self, model_name: Text, target_path: Text) -> Text:
        """Downloads a model that has been persisted to cloud storage.

        Downloaded model will be saved to the `target_path`.
        If `target_path` is a directory, the model will be downloaded to that directory.
        If `target_path` is a file, the model will be downloaded to that file.

        Args:
            model_name: The name of the model to retrieve.
            target_path: The path to which the model should be downloaded.
        """
        tar_name = model_name
        if not model_name.endswith(MODEL_ARCHIVE_EXTENSION):
            # ensure backward compatibility
            tar_name = self._tar_name(model_name)
        remote_object_path = self._create_file_key(tar_name)
        self._retrieve_tar(remote_object_path)

        target_tar_file_name = os.path.basename(tar_name)
        if os.path.isdir(target_path):
            target_path = os.path.join(target_path, target_tar_file_name)

        if not os.path.exists(target_path):
            structlogger.debug(
                "persistor.retrieve.copy_model",
                event_info=f"Copying model '{target_tar_file_name}' to "
                f"'{target_path}'.",
            )
            self._copy(target_tar_file_name, target_path)

        structlogger.debug(
            "persistor.retrieve.model_retrieved",
            event_info=f"Model retrieved and saved to '{target_path}'.",
        )
        return target_path

    @abc.abstractmethod
    def _retrieve_tar(self, filename: Text) -> None:
        """Downloads a model previously persisted to cloud storage."""
        raise NotImplementedError

    @abc.abstractmethod
    def _persist_tar(self, filekey: Text, tarname: Text) -> None:
        """Uploads a model persisted in the `target_dir` to cloud storage."""
        raise NotImplementedError

    def _compress(self, model_directory: Text, model_name: Text) -> Tuple[Text, Text]:
        """Creates a compressed archive and returns key and tar."""
        import tempfile

        dirpath = tempfile.mkdtemp()
        base_name = self._tar_name(model_name, include_extension=False)
        tar_name = shutil.make_archive(
            os.path.join(dirpath, base_name),
            "gztar",
            root_dir=model_directory,
            base_dir=".",
        )
        file_key = os.path.basename(tar_name)
        return file_key, tar_name

    @staticmethod
    def _tar_name(model_name: Text, include_extension: bool = True) -> Text:
        ext = f".{MODEL_ARCHIVE_EXTENSION}" if include_extension else ""
        return f"{model_name}{ext}"

    @staticmethod
    def _copy(compressed_path: Text, target_path: Text) -> None:
        shutil.copy2(compressed_path, target_path)

    @staticmethod
    def _create_file_key(model_path: str) -> Text:
        """Appends remote storage folders when provided to upload or retrieve file"""
        bucket_object_path = os.environ.get(REMOTE_STORAGE_PATH_ENV)

        # To keep the backward compatibility, if REMOTE_STORAGE_PATH is not provided,
        # the model_name (which might be a complete path) will be returned as it is.
        if bucket_object_path is None:
            return str(model_path)
        else:
            raise_warning(
                f"{REMOTE_STORAGE_PATH_ENV} is deprecated and will be "
                "removed in future versions. "
                "Please use the -m path/to/model.tar.gz option to "
                "specify the model path when loading a model."
                "Or use --output and --fixed-model-name to specify the "
                "output directory and the model name when saving a "
                "trained model to remote storage.",
            )

        file_key = os.path.basename(model_path)
        file_key = os.path.join(bucket_object_path, file_key)
        return file_key


class AWSPersistor(Persistor):
    """Store models on S3.

    Fetches them when needed, instead of storing them on the local disk.
    """

    def __init__(
        self,
        bucket_name: Text,
        endpoint_url: Optional[Text] = None,
        region_name: Optional[Text] = None,
    ) -> None:
        import boto3

        super().__init__()
        self.s3 = boto3.resource(
            "s3", endpoint_url=endpoint_url, region_name=region_name
        )
        self._ensure_bucket_exists(bucket_name, region_name)
        self.bucket_name = bucket_name
        self.bucket = self.s3.Bucket(bucket_name)

    def _ensure_bucket_exists(
        self, bucket_name: Text, region_name: Optional[Text] = None
    ) -> None:
        import botocore

        # noinspection PyUnresolvedReferences
        try:
            self.s3.meta.client.head_bucket(Bucket=bucket_name)
        except botocore.exceptions.ClientError as e:
            error_code = int(e.response["Error"]["Code"])
            if error_code == HTTP_STATUS_FORBIDDEN:
                log = (
                    f"Access to the specified bucket '{bucket_name}' is forbidden. "
                    "Please make sure you have the necessary "
                    "permission to access the bucket."
                )
                structlogger.error(
                    "aws_persistor.ensure_bucket_exists.bucket_access_forbidden",
                    bucket_name=bucket_name,
                    event_info=log,
                )
                raise RasaException(log)
            elif error_code == HTTP_STATUS_NOT_FOUND:
                log = (
                    f"The specified bucket '{bucket_name}' does not exist. "
                    "Please make sure to create the bucket first."
                )
                structlogger.error(
                    "aws_persistor.ensure_bucket_exists.bucket_not_found",
                    bucket_name=bucket_name,
                    event_info=log,
                )
                raise RasaException(log)

    def _persist_tar(self, file_key: Text, tar_path: Text) -> None:
        """Uploads a model persisted in the `target_dir` to s3."""
        structlogger.debug(
            "aws_persistor.persist_tar.uploading_model",
            event_info=f"Uploading tar archive {file_key} to "
            f"s3 bucket '{self.bucket_name}'.",
        )
        with open(tar_path, "rb") as f:
            self.s3.Object(self.bucket_name, file_key).put(Body=f)

    def _retrieve_tar(self, model_path: Text) -> None:
        """Downloads a model that has previously been persisted to s3."""
        tar_name = os.path.basename(model_path)
        with open(tar_name, "wb") as f:
            self.bucket.download_fileobj(model_path, f)


class GCSPersistor(Persistor):
    """Store models on Google Cloud Storage.

    Fetches them when needed, instead of storing them on the local disk.
    """

    def __init__(self, bucket_name: Text) -> None:
        """Initialise class with client and bucket."""
        # there are no type hints in this repo for now
        # https://github.com/googleapis/python-storage/issues/393
        from google.cloud import storage

        super().__init__()

        self.storage_client = storage.Client()
        self._ensure_bucket_exists(bucket_name)

        self.bucket_name = bucket_name
        self.bucket = self.storage_client.bucket(bucket_name)

    def _ensure_bucket_exists(self, bucket_name: Text) -> None:
        from google.cloud import exceptions

        try:
            self.storage_client.get_bucket(bucket_name)
        except exceptions.NotFound:
            log = (
                f"The specified bucket '{bucket_name}' does not exist. "
                "Please make sure to create the bucket first."
            )
            structlogger.error(
                "gcp_persistor.ensure_bucket_exists.bucket_not_found",
                bucket_name=bucket_name,
                event_info=log,
            )
            raise RasaException(log)
        except exceptions.Forbidden:
            log = (
                f"Access to the specified bucket '{bucket_name}' is forbidden. "
                "Please make sure you have the necessary "
                "permission to access the bucket. "
            )
            structlogger.error(
                "gcp_persistor.ensure_bucket_exists.bucket_access_forbidden",
                bucket_name=bucket_name,
                event_info=log,
            )
            raise RasaException(log)

    def _persist_tar(self, file_key: Text, tar_path: Text) -> None:
        """Uploads a model persisted in the `target_dir` to GCS."""
        blob = self.bucket.blob(file_key)
        blob.upload_from_filename(tar_path)

    def _retrieve_tar(self, target_filename: Text) -> None:
        """Downloads a model that has previously been persisted to GCS."""
        blob = self.bucket.blob(target_filename)
        blob.download_to_filename(os.path.basename(target_filename))


class AzurePersistor(Persistor):
    """Store models on Azure."""

    def __init__(
        self, azure_container: Text, azure_account_name: Text, azure_account_key: Text
    ) -> None:
        from azure.storage.blob import BlobServiceClient

        super().__init__()

        self.blob_service = BlobServiceClient(
            account_url=f"https://{azure_account_name}.blob.core.windows.net/",
            credential=azure_account_key,
        )
        self.container_name = azure_container
        self._ensure_container_exists()

    def _ensure_container_exists(self) -> None:
        if self._container_client().exists():
            pass
        else:
            log = (
                f"The specified container '{self.container_name}' does not exist."
                "Please make sure to create the container first."
            )
            structlogger.error(
                "azure_persistor.ensure_container_exists.container_not_found",
                container_name=self.container_name,
                event_info=log,
            )
            raise RasaException(log)

    def _container_client(self) -> "ContainerClient":
        return self.blob_service.get_container_client(self.container_name)

    def _persist_tar(self, file_key: Text, tar_path: Text) -> None:
        """Uploads a model persisted in the `target_dir` to Azure."""
        with open(tar_path, "rb") as data:
            self._container_client().upload_blob(
                name=file_key,
                data=data,
                # overwrite is set to True to keep in line with
                # how GCS and AWS APIs work this enables easy
                # updating of models in the cloud
                overwrite=True,
            )

    def _retrieve_tar(self, target_filename: Text) -> None:
        """Downloads a model that has previously been persisted to Azure."""
        blob_client = self._container_client().get_blob_client(target_filename)

        with open(os.path.basename(target_filename), "wb") as blob:
            download_stream = blob_client.download_blob()
            blob.write(download_stream.readall())



import argparse
import logging
from pathlib import Path
from typing import Any, Dict, List, Optional, Set, Tuple

import questionary
import structlog

import rasa.cli.utils
import rasa.shared.utils.cli
from rasa.shared.constants import (
    DEFAULT_DATA_PATH,
    DEFAULT_DOMAIN_PATH,
    DEFAULT_CONFIG_PATH,
    DEFAULT_ENDPOINTS_PATH,
)
from rasa.shared.core.domain import Domain
from rasa.shared.core.flows.yaml_flows_io import YamlFlowsWriter
from rasa.shared.importers.importer import TrainingDataImporter
from rasa.shared.utils.yaml import read_yaml
from rasa.studio import data_handler
from rasa.studio.config import StudioConfig
from rasa.studio.constants import (
    STUDIO_DOMAIN_FILENAME,
    STUDIO_FLOWS_FILENAME,
    STUDIO_NLU_FILENAME,
)
from rasa.studio.data_handler import (
    StudioDataHandler,
    import_data_from_studio,
)
from rasa.utils.mapper import RasaPrimitiveStorageMapper

logger = logging.getLogger(__name__)
structlogger = structlog.getLogger(__name__)


def _handle_file_overwrite(
    file_path: Optional[str], default_path: str, file_type: str
) -> Tuple[Optional[Path], bool]:
    """Handles the logic for determining whether to
    overwrite an existing file or create a new one.
    Works for config and endpoints at this moment

    Args:
        file_path (Optional[str]): The path to the file
                    provided by the user. Can be None.
        default_path (str): The default path to use if `file_path`
                    is None or invalid. Must be a file path.
        file_type (str): The type of the file (e.g., "config",
            "endpoints") for logging and messaging purposes.

    Returns:
        tuple[Optional[Path], bool]: A tuple containing the
                            resolved file path and a boolean
                        indicating whether to write the file.
    """
    file_already_exists = rasa.cli.utils.get_validated_path(
        file_path, file_type, default_path, none_is_valid=True
    )
    write_file = False
    path = None
    file_or_default_path = file_path or default_path

    if file_already_exists is None:
        path = Path(file_or_default_path)
        if path.is_dir():
            path = path / default_path
        return path, True

    if questionary.confirm(
        f"{file_type.capitalize()} file '{file_or_default_path}' "
        f"already exists. Do you want to overwrite it?"
    ).ask():
        write_file = True
        path = Path(file_or_default_path)
    return path, write_file


def _prepare_data_and_domain_paths(args: argparse.Namespace) -> Tuple[Path, List[Path]]:
    """Handles the logic for preparing the domain and data paths
    based on the provided arguments.

    Args:
        args (argparse.Namespace): The parsed arguments.

    Returns:
        tuple[Path, list[Path]]: A tuple containing the domain path
                        and a list of data paths.
    """
    # prepare domain
    domain_path = rasa.cli.utils.get_validated_path(
        args.domain, "domain", DEFAULT_DOMAIN_PATH, none_is_valid=True
    )
    domain_or_default_path = args.domain or DEFAULT_DOMAIN_PATH

    if domain_path is None:
        # If the path is None, use the provided domain path
        domain_path = Path(domain_or_default_path)
        domain_path.touch()

    if isinstance(domain_path, str):
        domain_path = Path(domain_path)

    if domain_path.is_file():
        if not args.overwrite:
            domain_path.unlink()
            domain_path.touch()

    if domain_path.is_dir():
        if not args.overwrite:
            domain_path = domain_path / STUDIO_DOMAIN_FILENAME
            domain_path.touch()

    # prepare data
    data_paths = []

    for f in args.data:
        data_path = rasa.cli.utils.get_validated_path(
            f, "data", DEFAULT_DATA_PATH, none_is_valid=True
        )

        if data_path is None:
            # If the path is None, use the default data path
            data_path = Path(f)
            data_path.mkdir(parents=True, exist_ok=True)
        else:
            data_path = Path(data_path)

        if data_path.is_file() or data_path.is_dir():
            # If it's a file, add it directly
            data_paths.append(data_path)
        else:
            # If it doesn't exist, create the directory
            data_path.mkdir(parents=True, exist_ok=True)
            data_paths.append(data_path)

    # Remove duplicates while preserving order
    data_paths = list(dict.fromkeys(data_paths))

    return domain_path, data_paths


def handle_download(args: argparse.Namespace) -> None:
    handler = StudioDataHandler(
        studio_config=StudioConfig.read_config(), assistant_name=args.assistant_name[0]
    )
    handler.request_all_data()

    domain_path, data_paths = _prepare_data_and_domain_paths(args)

    # handle config and endpoints
    config_path, write_config = _handle_file_overwrite(
        args.config, DEFAULT_CONFIG_PATH, "config"
    )
    endpoints_path, write_endpoints = _handle_file_overwrite(
        args.endpoints, DEFAULT_ENDPOINTS_PATH, "endpoints"
    )

    # generate log message if we write the config or endpoints
    message_parts = []

    config_path = config_path if write_config else None
    endpoints_path = endpoints_path if write_endpoints else None

    if config_path:
        config_data = handler.get_config()
        if not config_data:
            rasa.shared.utils.cli.print_error_and_exit("No config data found.")
        with open(config_path, "w") as f:
            f.write(config_data)
            message_parts.append(f"config to '{config_path}'")

    if endpoints_path:
        endpoints_data = handler.get_endpoints()
        if not endpoints_data:
            raise ValueError("No endpoints data found.")

        with open(endpoints_path, "w") as f:
            f.write(endpoints_data)
            message_parts.append(f"endpoints to '{endpoints_path}'")

    if message_parts:
        message = "Downloaded " + " and ".join(message_parts)
        structlogger.info("studio.download.config_endpoints", event_info=message)

    if not args.overwrite:
        _handle_download_no_overwrite(
            handler=handler,
            domain_path=domain_path,
            data_paths=data_paths,
        )
    else:
        _handle_download_with_overwrite(
            handler=handler,
            domain_path=domain_path,
            data_paths=data_paths,
        )


def _handle_download_no_overwrite(
    handler: StudioDataHandler,
    domain_path: Path,
    data_paths: List[Path],
) -> None:
    data_from_studio, data_original = import_data_from_studio(
        handler, domain_path, data_paths
    )

    if domain_path.is_dir():
        studio_domain_path = domain_path / STUDIO_DOMAIN_FILENAME
        new_domain_data = data_handler.combine_domains(
            data_from_studio.get_user_domain().as_dict(),
            data_original.get_user_domain().as_dict(),
        )
        studio_domain = Domain.from_dict(new_domain_data)
        if not studio_domain.is_empty():
            studio_domain.persist(studio_domain_path)
        else:
            logger.warning("No additional domain data found.")
    elif domain_path.is_file():
        domain_merged = data_original.get_user_domain().merge(
            data_from_studio.get_user_domain()
        )
        domain_merged.persist(domain_path)

    if len(data_paths) == 1 and data_paths[0].is_file():
        data_path = data_paths[0]
        if handler.has_nlu():
            data_nlu = data_original.get_nlu_data().merge(
                data_from_studio.get_nlu_data()
            )
            data_nlu.persist_nlu(data_path)
        if handler.has_flows():
            data_flows = data_original.get_user_flows().merge(
                data_from_studio.get_user_flows()
            )
            YamlFlowsWriter.dump(data_flows.underlying_flows, data_path)

    elif len(data_paths) == 1 and data_paths[0].is_dir():
        if handler.has_nlu():
            data_path = data_paths[0] / Path(STUDIO_NLU_FILENAME)
            _persist_nlu_diff(data_original, data_from_studio, data_path)
        if handler.has_flows():
            data_path = data_paths[0] / Path(STUDIO_FLOWS_FILENAME)
            _persist_flows_diff(data_original, data_from_studio, data_path)
    else:
        if data_paths[0].is_dir():
            logger.info(f"Saving data to {data_paths[0]}.")
            data_path = data_paths[0] / Path(STUDIO_NLU_FILENAME)
        else:
            logger.info(f"Saving data to {STUDIO_NLU_FILENAME}.")
            data_path = Path(STUDIO_NLU_FILENAME)
        if handler.has_nlu():
            _persist_nlu_diff(data_original, data_from_studio, data_path)
        if handler.has_flows():
            _persist_flows_diff(data_original, data_from_studio, data_path)


def _persist_nlu_diff(
    data_original: TrainingDataImporter,
    data_from_studio: TrainingDataImporter,
    data_path: Path,
) -> None:
    """Creates a new nlu file from the diff of original and studio data."""
    new_nlu_data = data_handler.create_new_nlu_from_diff(
        read_yaml(data_from_studio.get_nlu_data().nlu_as_yaml()),
        read_yaml(data_original.get_nlu_data().nlu_as_yaml()),
    )
    if new_nlu_data["nlu"]:
        pretty_write_nlu_yaml(new_nlu_data, data_path)
    else:
        logger.warning("No additional nlu data found.")


def _persist_flows_diff(
    data_original: TrainingDataImporter,
    data_from_studio: TrainingDataImporter,
    data_path: Path,
) -> None:
    """Creates a new flows file from the diff of original and studio data."""
    new_flows_data = data_handler.create_new_flows_from_diff(
        data_from_studio.get_user_flows().underlying_flows,
        data_original.get_user_flows().underlying_flows,
    )
    if new_flows_data:
        YamlFlowsWriter.dump(new_flows_data, data_path)
    else:
        logger.warning("No additional flows data found.")


def pretty_write_nlu_yaml(data: Dict, file: Path) -> None:
    """Writes the nlu yaml in a pretty way.

    Args:
        data: The data to write.
        file: The file to write to.
    """
    from ruamel import yaml
    from ruamel.yaml.scalarstring import LiteralScalarString

    dumper = yaml.YAML()
    for item in data["nlu"]:
        if item.get("examples"):
            item["examples"] = LiteralScalarString(item["examples"])

    with file.open("w", encoding="utf-8") as outfile:
        dumper.dump(data, outfile)


def _handle_download_with_overwrite(
    handler: StudioDataHandler,
    domain_path: Path,
    data_paths: List[Path],
) -> None:
    data_from_studio, data_original = import_data_from_studio(
        handler, domain_path, data_paths
    )
    mapper = RasaPrimitiveStorageMapper(
        domain_path=domain_path, training_data_paths=data_paths
    )

    if domain_path.is_file():
        domain_merged = data_from_studio.get_user_domain().merge(
            data_original.get_user_domain()
        )
        domain_merged.persist(domain_path)
    elif domain_path.is_dir():
        default = domain_path / Path(STUDIO_DOMAIN_FILENAME)
        studio_domain = data_from_studio.get_user_domain()

        paths = get_domain_path(domain_path, data_from_studio, mapper)

        _persist_domain_part(
            {"intents": studio_domain.as_dict().get("intents", [])},
            default,
            paths["intent_path"],
        )
        _persist_domain_part(
            {"entities": studio_domain.as_dict().get("entities", [])},
            default,
            paths["entities_path"],
        )
        _persist_domain_part(
            {"actions": studio_domain.as_dict().get("actions", [])},
            default,
            paths["action_path"],
        )
        _persist_domain_part(
            {"slots": studio_domain.as_dict().get("slots", {})},
            default,
            paths["slot_path"],
        )
        _persist_domain_part(
            {"responses": studio_domain.as_dict().get("responses", {})},
            default,
            paths["response_path"],
        )

    if len(data_paths) == 1 and data_paths[0].is_file():
        if handler.has_nlu():
            nlu_data_merged = data_from_studio.get_nlu_data().merge(
                data_original.get_nlu_data()
            )
            nlu_data_merged.persist_nlu(data_paths[0])
        if handler.has_flows():
            flows_data_merged = data_from_studio.get_user_flows().merge(
                data_original.get_user_flows()
            )
            YamlFlowsWriter.dump(flows_data_merged.underlying_flows, data_paths[0])
    elif len(data_paths) == 1 and data_paths[0].is_dir():
        data_path = data_paths[0]
        paths = get_training_path(data_path, data_original, mapper)
        if handler.has_nlu():
            nlu_data = data_from_studio.get_nlu_data()
            if paths["nlu_path"].exists():
                nlu_file = TrainingDataImporter.load_from_dict(
                    training_data_paths=[str(paths["nlu_path"])]
                )
                nlu_data = nlu_data.merge(nlu_file.get_nlu_data())
            pretty_write_nlu_yaml(read_yaml(nlu_data.nlu_as_yaml()), paths["nlu_path"])
        if handler.has_flows():
            flows_data = data_from_studio.get_user_flows()
            if paths["flow_path"].exists():
                flows_file = TrainingDataImporter.load_from_dict(
                    training_data_paths=[str(paths["flow_path"])]
                )
                flows_data = flows_data.merge(flows_file.get_user_flows())
            YamlFlowsWriter.dump(flows_data.underlying_flows, paths["flow_path"])
    else:
        #  TODO: we are not handling the case of multiple data paths?
        raise NotImplementedError("Multiple data paths are not supported yet.")


def _persist_domain_part(
    domain_part: Dict[str, Any], default: Path, path: Optional[Path]
) -> None:
    domain = Domain.from_dict(domain_part)
    if (path is not None) and path.exists():
        domain = Domain.from_file(path).merge(domain, override=True)
        domain.persist(path)
    else:
        domain.persist(default)


def get_training_path(
    path: Path,
    data_original: TrainingDataImporter,
    mapper: RasaPrimitiveStorageMapper,
) -> Dict[str, Path]:
    nlu_paths = set()
    flow_paths = set()
    for intent in data_original.get_nlu_data().intents:
        for path in mapper.get_file(intent, "intents").get("training", []):
            nlu_paths.add(path)
    flows = [flow.id for flow in data_original.get_user_flows().underlying_flows]
    for flow in flows:
        for path in mapper.get_file(flow, "flows").get("training", []):
            flow_paths.add(path)

    return {
        "nlu_path": _select_path(nlu_paths, "nlu", path, STUDIO_NLU_FILENAME),
        "flow_path": _select_path(flow_paths, "flows", path, STUDIO_FLOWS_FILENAME),
    }


def get_domain_path(
    domain_path: Path,
    data_from_studio: TrainingDataImporter,
    mapper: RasaPrimitiveStorageMapper,
) -> Dict[str, Path]:
    intent_paths = set()
    entities_paths = set()
    slot_paths = set()
    action_paths = set()
    response_paths = set()
    # nlu-based
    for intent in data_from_studio.get_domain().intents:
        for path in mapper.get_file(intent, "intents").get("domain", []):
            intent_paths.add(path)
    for entity in data_from_studio.get_domain().entities:
        for path in mapper.get_file(entity, "entities").get("domain", []):
            entities_paths.add(path)
    # flow-based
    slot_list = [slot.name for slot in data_from_studio.get_domain().slots]
    for slot in slot_list:
        for path in mapper.get_file(slot, "slots").get("domain", []):
            slot_paths.add(path)
    for action in data_from_studio.get_domain().action_names_or_texts:
        for path in mapper.get_file(action, "actions").get("domain", []):
            action_paths.add(path)
    for response in data_from_studio.get_domain().responses:
        for path in mapper.get_file(response, "responses").get("domain", []):
            response_paths.add(path)

    return {
        "intent_path": _select_path(
            intent_paths, "intents", domain_path, STUDIO_DOMAIN_FILENAME
        ),
        "entities_path": _select_path(
            entities_paths, "entities", domain_path, STUDIO_DOMAIN_FILENAME
        ),
        "slot_path": _select_path(
            slot_paths, "slots", domain_path, STUDIO_DOMAIN_FILENAME
        ),
        "action_path": _select_path(
            action_paths, "actions", domain_path, STUDIO_DOMAIN_FILENAME
        ),
        "response_path": _select_path(
            response_paths, "responses", domain_path, STUDIO_DOMAIN_FILENAME
        ),
    }


def _select_path(
    paths: Set[Path], primitive_type: str, default_path: Path, default: str
) -> Path:
    if len(paths) == 1:
        path = paths.pop()
    elif len(paths) > 1:
        path = paths.pop()
        logger.warning(
            f"Saving {primitive_type} to {path}."
            f"Please keep Studio related {primitive_type} in a single file."
        )
    else:  # no path in paths
        path = default_path / Path(default)
        logger.info(f"Saving {primitive_type} to {path}.")
    return path



import os
from typing import Optional, Text, cast

from rasa.shared.exceptions import RasaException
from rasa.utils.endpoints import EndpointConfig, read_endpoint_config

from rasa.core.secrets_manager.constants import (
    SECRET_MANAGER_ENV_NAME,
    VAULT_DEFAULT_RASA_SECRETS_PATH,
    VAULT_ENDPOINT_NAMESPACE_LABEL,
    VAULT_ENDPOINT_SECRETS_PATH_LABEL,
    VAULT_ENDPOINT_TRANSIT_MOUNT_POINT_LABEL,
    VAULT_NAMESPACE_ENV_NAME,
    VAULT_RASA_SECRETS_PATH_ENV_NAME,
    VAULT_SECRET_MANAGER_NAME,
    VAULT_TOKEN_ENV_NAME,
    VAULT_TRANSIT_MOUNT_POINT_ENV_NAME,
    VAULT_URL_ENV_NAME,
)
from rasa.core.secrets_manager.secret_manager import (
    SecretManagerConfig,
    SecretsManager,
    SecretsManagerProvider,
)
from rasa.core.secrets_manager.vault import (
    VaultSecretManagerConfig,
    VaultSecretManagerNonStrictConfig,
    VaultSecretsManager,
)


def create(config: SecretManagerConfig) -> Optional[SecretsManager]:
    """Create a SecretsManager based on the configuration.

    Args:
        config: SecretManagerConfig

    Returns:
        SecretsManager created based on the configuration
    """
    secret_manager = None

    if config.secret_manager_type == VAULT_SECRET_MANAGER_NAME:
        vault_config = cast(VaultSecretManagerConfig, config)
        secret_manager = VaultSecretsManager(
            host=vault_config.url,
            token=vault_config.token,
            transit_mount_point=vault_config.transit_mount_point,
            secrets_path=vault_config.secrets_path,
            namespace=vault_config.namespace,
        )

    return secret_manager


def read_vault_endpoint_config(
    endpoints_file: Optional[Text],
) -> Optional[VaultSecretManagerNonStrictConfig]:
    """Read endpoints file to discover vault config.

    Args:
        endpoints_file: Path to the endpoints file

    Returns:
        A configuration for the vault secret manager
    """
    if endpoints_file is None:
        return None

    endpoint_config = read_endpoint_config(
        filename=endpoints_file, endpoint_type="secrets_manager"
    )

    if endpoint_config:
        url = endpoint_config.url
        token = endpoint_config.token
        transit_mount_point = endpoint_config.kwargs.get(
            VAULT_ENDPOINT_TRANSIT_MOUNT_POINT_LABEL
        )
        secrets_path = endpoint_config.kwargs.get(VAULT_ENDPOINT_SECRETS_PATH_LABEL)
        namespace = endpoint_config.kwargs.get(VAULT_ENDPOINT_NAMESPACE_LABEL)

        return VaultSecretManagerNonStrictConfig(
            url=url,
            token=token,
            transit_mount_point=transit_mount_point,
            secrets_path=secrets_path or VAULT_DEFAULT_RASA_SECRETS_PATH,
            namespace=namespace,
        )

    return None


def read_vault_env_vars() -> VaultSecretManagerNonStrictConfig:
    """Read environment variables to discover vault config.

    Returns:
        A configuration for the vault secret manager
    """
    url = os.getenv(VAULT_URL_ENV_NAME)
    token = os.getenv(VAULT_TOKEN_ENV_NAME)
    transit_mount_point = os.getenv(VAULT_TRANSIT_MOUNT_POINT_ENV_NAME)
    secrets_path = os.getenv(VAULT_RASA_SECRETS_PATH_ENV_NAME)
    namespace = os.getenv(VAULT_NAMESPACE_ENV_NAME)

    return VaultSecretManagerNonStrictConfig(
        url=url,
        token=token,
        transit_mount_point=transit_mount_point,
        secrets_path=secrets_path,
        namespace=namespace,
    )


def read_vault_config(
    endpoints_file: Optional[Text],
) -> Optional[VaultSecretManagerConfig]:
    """Read endpoints file to discover vault config.

    Args:
        endpoints_file: Path to the endpoints file

    Returns:
        A configuration for the vault secret manager
    """
    env_config = read_vault_env_vars()
    endpoint_config = read_vault_endpoint_config(endpoints_file)

    if env_config.is_valid() and endpoint_config is None:
        return VaultSecretManagerConfig(
            **env_config.__dict__,
        )

    if endpoint_config is not None:
        vault_config = env_config.merge(endpoint_config)

        if vault_config.is_empty():
            return None

        if vault_config.is_valid():
            return VaultSecretManagerConfig(
                **vault_config.__dict__,
            )

    raise RasaException(
        f"Cannot start Vault secret manager based on configuration from env vars "
        f"{VAULT_URL_ENV_NAME} = {env_config.url}, "
        f"{VAULT_TOKEN_ENV_NAME} = {env_config.token}, "
        f"{VAULT_RASA_SECRETS_PATH_ENV_NAME} = {env_config.secrets_path}, "
        f"{VAULT_TRANSIT_MOUNT_POINT_ENV_NAME} = {env_config.transit_mount_point}. "
        f"{VAULT_NAMESPACE_ENV_NAME} = {env_config.namespace}. "
    )


def read_secret_manager_from_endpoint_config(
    endpoints_file: Text,
) -> Optional[EndpointConfig]:
    """Read endpoints file to discover secret manager.

    Args:
        endpoints_file: Path to the endpoints file

    Returns:
        A secret manager
    """
    secret_manager_config = read_endpoint_config(
        filename=endpoints_file, endpoint_type="secrets_manager"
    )
    if secret_manager_config is None:
        return None

    return secret_manager_config


def read_secret_manager_config(
    endpoints_file: Optional[Text],
) -> Optional[SecretManagerConfig]:
    """Read endpoints file to discover secret manager config.

    If the secret manager is configured, set the environment variables required
    to connect to the secret manager.
    The above steps are skipped if the environment variables are already set.

    Args:
        endpoints_file: Path to the endpoints file

    Returns:
        A configuration for the secret manager
    """
    secret_manager_name = os.getenv(SECRET_MANAGER_ENV_NAME)

    if secret_manager_name is None:
        if endpoints_file is None:
            return None

        secret_manager_config = read_secret_manager_from_endpoint_config(
            endpoints_file=endpoints_file
        )
        if secret_manager_config is None:
            return None

        secret_manager_name = secret_manager_config.type

    if secret_manager_name == VAULT_SECRET_MANAGER_NAME:
        return read_vault_config(endpoints_file)

    return None


def load_secret_manager(endpoints_file: Optional[Text]) -> Optional[SecretsManager]:
    """Create secret manager based on the configuration in the endpoints file.

    Args:
        endpoints_file: Path to the endpoints file.

    Returns:
        The secret manager or `None` if no secret manager is configured.
    """
    secret_manager_config = read_secret_manager_config(endpoints_file=endpoints_file)

    if (
        secret_manager_config is None
        or secret_manager_config.secret_manager_type is None
    ):
        return None

    provider = SecretsManagerProvider()
    secret_manager = create(secret_manager_config)

    if secret_manager is not None:
        provider.register_secret_manager(secret_manager)

    return secret_manager



from typing import Dict, Any, Optional, Text

import structlog
from deprecated import deprecated  # type: ignore[import]

from rasa.dialogue_understanding.generator.single_step.single_step_llm_command_generator import (  # noqa: E501
    SingleStepLLMCommandGenerator,
)
from rasa.engine.recipes.default_recipe import DefaultV1Recipe
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.shared.exceptions import ProviderClientAPIException
from rasa.shared.utils.io import raise_deprecation_warning

structlogger = structlog.get_logger()


@DefaultV1Recipe.register(
    [
        DefaultV1Recipe.ComponentType.COMMAND_GENERATOR,
    ],
    is_trainable=True,
)
@deprecated(
    reason=(
        "The LLMCommandGenerator is deprecated and will be removed in Rasa 4.0.0. "
        "Please use use SingleStepLLMCommandGenerator instead."
    )
)
class LLMCommandGenerator(SingleStepLLMCommandGenerator):
    def __init__(
        self,
        config: Dict[str, Any],
        model_storage: ModelStorage,
        resource: Resource,
        prompt_template: Optional[Text] = None,
        **kwargs: Any,
    ) -> None:
        raise_deprecation_warning(
            message="`LLMCommandGenerator` has been renamed "
            "to `SingleStepLLMCommandGenerator`."
            "Support for the former name `LLMCommandGenerator` "
            "will be removed in Rasa `4.0.0`."
            "Please modify your assistant's configuration to "
            "use `SingleStepLLMCommandGenerator` instead."
        )

        super().__init__(
            config,
            model_storage,
            resource,
            prompt_template=prompt_template,
            **kwargs,
        )

    async def invoke_llm(self, prompt: Text) -> Optional[Text]:
        try:
            return await super().invoke_llm(prompt)
        except ProviderClientAPIException:
            # Returning 'None' in case of a ProviderClientAPIException ensures
            # backward compatibility with previous versions. In earlier
            # implementations, 'invoke_llm' was expected to return 'None' when
            # encountering API-related issues. This practice is maintained here
            # to prevent breaking changes in how exceptions are handled and
            # propagated in existing integrations, where callers might not yet
            # be prepared to manage exceptions directly.
            return None



from __future__ import annotations
import dataclasses

from typing import Dict, List, Text, Any

from rasa.engine.graph import GraphComponent, ExecutionContext
from rasa.engine.storage.resource import Resource
from rasa.engine.storage.storage import ModelStorage
from rasa.shared.core.domain import Domain
import rasa.shared.utils.io


@dataclasses.dataclass
class Responses:
    """Holds the responses of the domain."""

    data: Dict[Text, List[Dict[Text, Any]]]

    def fingerprint(self) -> Text:
        """Returns a fingerprint of the responses."""
        return rasa.shared.utils.io.get_dictionary_fingerprint(self.data)

    def get(self, key: Text, default: Any) -> Any:
        """Returns the value for the given key."""
        return self.data.get(key, default)


class ResponsesProvider(GraphComponent):
    """Provides responses during training and inference time."""

    @classmethod
    def create(
        cls,
        config: Dict[Text, Any],
        model_storage: ModelStorage,
        resource: Resource,
        execution_context: ExecutionContext,
    ) -> ResponsesProvider:
        """Creates component (see parent class for full docstring)."""
        return cls()

    def provide(self, domain: Domain) -> Responses:
        """Returns the responses from the given domain."""
        return Responses(data=domain.responses)



import json
import logging
import time
from collections import deque
from typing import Deque, Optional, Text

from rasa.core.lock import Ticket, TicketLock
from rasa.core.lock_store import (
    DEFAULT_SOCKET_TIMEOUT_IN_SECONDS,
    LOCK_LIFETIME,
    LockError,
    LockStore,
)
from rasa.utils.endpoints import EndpointConfig

DEFAULT_REDIS_DB = 1

DEFAULT_PORT = 6379

DEFAULT_HOSTNAME = "localhost"

logger = logging.getLogger(__name__)

LAST_ISSUED_TICKET_NUMBER_SUFFIX = "last_issued_ticket_number"

DEFAULT_CONCURRENT_REDIS_LOCK_STORE_KEY_PREFIX = "concurrent_lock:"


class ConcurrentTicketLock(TicketLock):
    def concurrent_issue_ticket(self, lifetime: float, ticket_number: int) -> int:
        """Issue a new ticket and return its number.

        Args:
            lifetime: The length of time after which the ticket expires.
            ticket_number: Integer representing the ticket number.

        Returns:
            ticket_number integer.

        """
        ticket = Ticket(ticket_number, time.time() + lifetime)
        self.tickets.append(ticket)

        return ticket_number


class ConcurrentRedisLockStore(LockStore):
    """Concurrent implementation of a Redis store for ticket locks."""

    def __init__(
        self,
        endpoint_config: EndpointConfig,
    ) -> None:
        """Create a lock store which uses Redis for persistence.

        Args:
            endpoint_config: The endpointConfig defined the lock store.

            Can contain the following properties:
            host - The host of the redis server.
            port - The port of the redis server.
            db - The name of the database within Redis which should be used by Rasa
                Open Source.
            username - The username which should be used for authentication with the
                Redis database.
            password - The password which should be used for authentication with the
                Redis database.
            use_ssl - `True` if SSL should be used for the connection to Redis.
            ssl_certfile - Path to the SSL certificate file.
            ssl_keyfile - Path to the SSL private key file.
            ssl_ca_certs - Path to the SSL CA certificate file.
            key_prefix - prefix to prepend to all keys used by the lock store. Must be
                alphanumeric.
            socket_timeout - Timeout in seconds after which an exception will be raised
                in case Redis doesn't respond within `socket_timeout` seconds.
        """
        import redis

        host = endpoint_config.kwargs.get("host", DEFAULT_HOSTNAME)
        port = endpoint_config.kwargs.get("port", DEFAULT_PORT)
        db = endpoint_config.kwargs.get("db", DEFAULT_REDIS_DB)
        username = endpoint_config.kwargs.get("username")
        password = endpoint_config.kwargs.get("password")
        use_ssl = bool(endpoint_config.kwargs.get("use_ssl", False))
        ssl_certfile = endpoint_config.kwargs.get("ssl_certfile")
        ssl_keyfile = endpoint_config.kwargs.get("ssl_keyfile")
        ssl_ca_certs = endpoint_config.kwargs.get("ssl_ca_certs")
        key_prefix = endpoint_config.kwargs.get("key_prefix")
        socket_timeout = endpoint_config.kwargs.get(
            "socket_timeout", DEFAULT_SOCKET_TIMEOUT_IN_SECONDS
        )

        self.red = redis.StrictRedis(
            host=host,
            port=int(port),
            db=int(db),
            username=username,
            password=password,
            ssl=use_ssl,
            ssl_certfile=ssl_certfile,
            ssl_keyfile=ssl_keyfile,
            ssl_ca_certs=ssl_ca_certs,
            socket_timeout=socket_timeout,
        )

        self.key_prefix = DEFAULT_CONCURRENT_REDIS_LOCK_STORE_KEY_PREFIX
        if key_prefix:
            logger.debug(f"Setting non-default redis key prefix: '{key_prefix}'.")
            self._set_key_prefix(key_prefix)

        super().__init__()

    def _set_key_prefix(self, key_prefix: Text) -> None:
        if isinstance(key_prefix, str) and key_prefix.isalnum():
            self.key_prefix = (
                key_prefix + ":" + DEFAULT_CONCURRENT_REDIS_LOCK_STORE_KEY_PREFIX
            )
        else:
            logger.warning(
                f"Omitting provided non-alphanumeric redis key prefix: '{key_prefix}'. "
                f"Using default '{self.key_prefix}' instead."
            )

    def issue_ticket(
        self, conversation_id: Text, lock_lifetime: float = LOCK_LIFETIME
    ) -> int:
        """Issue new ticket for a conversation.

        It's configured with `lock_lifetime` and associated with `conversation_id`.
        Creates a new lock if none is found.
        """
        logger.debug(f"Issuing ticket for conversation '{conversation_id}'.")
        try:
            lock = self.get_or_create_lock(conversation_id)
            lock.remove_expired_tickets()
            ticket_number = self.increment_ticket_number(lock)
            ticket = lock.concurrent_issue_ticket(lock_lifetime, ticket_number)  # type: ignore[attr-defined]
            self.save_lock(lock)

            return ticket
        except Exception as e:
            raise LockError(f"Error while acquiring lock. Error:\n{e}")

    def get_lock(self, conversation_id: Text) -> Optional[ConcurrentTicketLock]:
        """Retrieves lock (see parent docstring for more information)."""
        tickets: Deque[Ticket] = deque()

        pattern = self.key_prefix + conversation_id + ":" + "[0-9]*"
        redis_keys = self.red.keys(pattern)

        for key in redis_keys:
            serialised_ticket = self.red.get(key)
            if serialised_ticket:
                ticket = Ticket.from_dict(json.loads(serialised_ticket))
                tickets.appendleft(ticket)

        tickets = deque(sorted(tickets, key=lambda x: x.number))

        return ConcurrentTicketLock(conversation_id, tickets)

    def delete_lock(self, conversation_id: Text) -> None:
        """Deletes lock for conversation ID."""
        pattern = self.key_prefix + conversation_id + ":*"
        redis_keys = self.red.keys(pattern)

        if not redis_keys:
            logger.debug(
                f"The lock store does not contain any key-value "
                f"items for conversation '{conversation_id}'."
            )
            return None

        deletion_successful = self.red.delete(*redis_keys)
        if deletion_successful == 0:
            self._log_deletion(conversation_id, False)
        else:
            self._log_deletion(conversation_id, True)

    def save_lock(self, lock: TicketLock) -> None:
        """Commit individual tickets and last issued ticket number to storage."""
        last_issued_ticket = lock.tickets[-1]
        serialised_ticket = last_issued_ticket.dumps()
        key = (
            self.key_prefix
            + lock.conversation_id
            + ":"
            + str(last_issued_ticket.number)
        )
        self.red.set(
            name=key, value=serialised_ticket, ex=int(last_issued_ticket.expires)
        )

    def increment_ticket_number(self, lock: TicketLock) -> int:
        """Uses Redis atomic transaction to increment ticket number."""
        last_issued_key = (
            self.key_prefix
            + lock.conversation_id
            + ":"
            + LAST_ISSUED_TICKET_NUMBER_SUFFIX
        )

        return self.red.incr(name=last_issued_key)

    def finish_serving(self, conversation_id: Text, ticket_number: int) -> None:
        """Finish serving ticket with `ticket_number` for `conversation_id`.

        Removes ticket from storage.
        """
        ticket_key = self.key_prefix + conversation_id + ":" + str(ticket_number)
        self.red.delete(ticket_key)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List
from rasa.dialogue_understanding.commands import Command
from rasa.shared.core.events import Event
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.trackers import DialogueStateTracker


@dataclass
class ChangeFlowCommand(Command):
    """A command to indicate a change of flows was requested by the command
    generator."""

    @classmethod
    def command(cls) -> str:
        """Returns the command type."""
        return "change_flow"

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> ChangeFlowCommand:
        """Converts the dictionary to a command.

        Returns:
            The converted dictionary.
        """
        return ChangeFlowCommand()

    def run_command_on_tracker(
        self,
        tracker: DialogueStateTracker,
        all_flows: FlowsList,
        original_tracker: DialogueStateTracker,
    ) -> List[Event]:
        # the change flow command is not actually pushing anything to the tracker,
        # but it is predicted by the MultiStepLLMCommandGenerator and used internally
        return []

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, ChangeFlowCommand)

    def __hash__(self) -> int:
        return hash(self.command())



import copy
import logging
from pathlib import Path
from typing import Any, Dict, Text, Type, Union

from rasa.engine.caching import TrainingCache
from rasa.engine.graph import ExecutionContext, GraphSchema, GraphModelConfiguration
from rasa.engine.constants import PLACEHOLDER_IMPORTER
from rasa.engine.runner.interface import GraphRunner
from rasa.engine.storage.storage import ModelStorage, ModelMetadata
from rasa.engine.training.components import (
    PrecomputedValueProvider,
    FingerprintComponent,
    FingerprintStatus,
)
from rasa.engine.training.hooks import TrainingHook, LoggingHook
from rasa.shared.importers.importer import TrainingDataImporter

logger = logging.getLogger(__name__)


class GraphTrainer:
    """Trains a model using a graph schema."""

    def __init__(
        self,
        model_storage: ModelStorage,
        cache: TrainingCache,
        graph_runner_class: Type[GraphRunner],
    ) -> None:
        """Initializes a `GraphTrainer`.

        Args:
            model_storage: Storage which graph components can use to persist and load.
                Also used for packaging the trained model.
            cache: Cache used to store fingerprints and outputs.
            graph_runner_class: The class to instantiate the runner from.
        """
        self._model_storage = model_storage
        self._cache = cache
        self._graph_runner_class = graph_runner_class

    async def train(
        self,
        model_configuration: GraphModelConfiguration,
        importer: TrainingDataImporter,
        output_filename: Path,
        force_retraining: bool = False,
        is_finetuning: bool = False,
    ) -> ModelMetadata:
        """Trains and packages a model and returns the prediction graph runner.

        Args:
            model_configuration: The model configuration (schemas, language, etc.)
            importer: The importer which provides the training data for the training.
            output_filename: The location to save the packaged model.
            force_retraining: If `True` then the cache is skipped and all components
                are retrained.
            is_finetuning: `True` if we want to finetune the model.

        Returns:
            The metadata describing the trained model.
        """
        logger.debug("Starting training.")

        # Retrieve the domain for the model metadata right at the start.
        # This avoids that something during the graph runs mutates it.
        domain = copy.deepcopy(importer.get_domain())

        if force_retraining:
            logger.debug(
                "Skip fingerprint run as a full training of the model was enforced."
            )
            pruned_training_schema = model_configuration.train_schema
        else:
            fingerprint_run_outputs = await self.fingerprint(
                model_configuration.train_schema,
                importer=importer,
                is_finetuning=is_finetuning,
            )
            pruned_training_schema = self._prune_schema(
                model_configuration.train_schema, fingerprint_run_outputs
            )

        hooks = [
            LoggingHook(pruned_schema=pruned_training_schema),
            TrainingHook(
                cache=self._cache,
                model_storage=self._model_storage,
                pruned_schema=pruned_training_schema,
            ),
        ]

        graph_runner = self._graph_runner_class.create(
            graph_schema=pruned_training_schema,
            model_storage=self._model_storage,
            execution_context=ExecutionContext(
                graph_schema=model_configuration.train_schema,
                is_finetuning=is_finetuning,
            ),
            hooks=hooks,
        )

        logger.debug("Running the pruned train graph with real node execution.")

        await graph_runner.run(inputs={PLACEHOLDER_IMPORTER: importer})

        return self._model_storage.create_model_package(
            output_filename, model_configuration, domain
        )

    async def fingerprint(
        self,
        train_schema: GraphSchema,
        importer: TrainingDataImporter,
        is_finetuning: bool = False,
    ) -> Dict[Text, Union[FingerprintStatus, Any]]:
        """Runs the graph using fingerprints to determine which nodes need to re-run.

        Nodes which have a matching fingerprint key in the cache can either be removed
        entirely from the graph, or replaced with a cached value if their output is
        needed by descendent nodes.

        Args:
            train_schema: The train graph schema that will be run in fingerprint mode.
            importer: The importer which provides the training data for the training.
            is_finetuning: `True` if we want to finetune the model.

        Returns:
            Mapping of node names to fingerprint results.
        """
        fingerprint_schema = self._create_fingerprint_schema(train_schema)

        fingerprint_graph_runner = self._graph_runner_class.create(
            graph_schema=fingerprint_schema,
            model_storage=self._model_storage,
            execution_context=ExecutionContext(
                graph_schema=train_schema, is_finetuning=is_finetuning
            ),
        )

        logger.debug("Running the train graph in fingerprint mode.")
        return await fingerprint_graph_runner.run(
            inputs={PLACEHOLDER_IMPORTER: importer}
        )

    def _create_fingerprint_schema(self, train_schema: GraphSchema) -> GraphSchema:
        fingerprint_schema = copy.deepcopy(train_schema)
        for node_name, schema_node in fingerprint_schema.nodes.items():
            # We make every node a target so that `graph_runner.run(...)` returns
            # the output for each node. We need the output of each node
            # to decide which nodes we can prune.
            schema_node.is_target = True

            # We do not replace the input nodes as we need an up-to-date fingerprint of
            # any input data to the graph. This means we can prune according to what
            # has actually changed.
            if not schema_node.is_input:
                FingerprintComponent.replace_schema_node(schema_node, self._cache)
        return fingerprint_schema

    def _prune_schema(
        self,
        schema: GraphSchema,
        fingerprint_run_outputs: Dict[Text, Union[FingerprintStatus, Any]],
    ) -> GraphSchema:
        """Uses the fingerprint statuses to prune the graph schema.

        Walks the graph starting at each target node. If a node has a cache hit we
        replace it with a `PrecomputedValueProvider` and remove its input dependencies.
        At the end, any node that is not an ancestor of a target node will be pruned
        when we call `minimal_graph_schema()`.

        Args:
            schema: The graph to prune.
            fingerprint_run_outputs: Node outputs from the fingerprint run as a mapping
                from node name to output.

        Returns:
            The pruned schema.
        """
        pruned_schema = copy.deepcopy(schema)
        target_node_names = pruned_schema.target_names

        for target_node_name in target_node_names:
            self._walk_and_prune(
                pruned_schema, target_node_name, fingerprint_run_outputs
            )

        return pruned_schema.minimal_graph_schema()

    def _walk_and_prune(
        self,
        schema: GraphSchema,
        current_node_name: Text,
        fingerprint_run_outputs: Dict[Text, Union[FingerprintStatus, Any]],
    ) -> None:
        """Recursively walks backwards though a graph checking the status of each node.

        If node has a fingerprint key hit then we check if there is a cached output.
        If there is a cached output we will replace the node with a
        `PrecomputedValueProvider` and remove all its dependencies (`.needs`). If
        there is not a fingerprint key hit, or there is no cached output, the node is
        left untouched and will be executed again next run unless it is no longer the
        ancestor of a target node.

        Args:
            schema: The graph we are currently walking.
            current_node_name: The current node on the walk.
            fingerprint_run_outputs: The fingerprint statuses of every node as a mapping
                from node name to status.
        """
        fingerprint_run_output = fingerprint_run_outputs[current_node_name]
        node = schema.nodes[current_node_name]

        # If we have replaced this node with a `PrecomputedValueProvider` we have
        # already visited this node. A `PrecomputedValueProvider` is updated to have
        # no parent nodes, so
        # we can end the walk here.
        if node.uses == PrecomputedValueProvider:
            return

        # If the output was a `FingerprintStatus` we must check the cache and status.
        if isinstance(fingerprint_run_output, FingerprintStatus):
            # If there is a fingerprint key hit we can potentially use a cached output.
            if fingerprint_run_output.is_hit:
                output_result = self._cache.get_cached_result(
                    output_fingerprint_key=fingerprint_run_output.output_fingerprint,
                    node_name=current_node_name,
                    model_storage=self._model_storage,
                )
                if output_result:
                    logger.debug(
                        f"Updating '{current_node_name}' to use a "
                        f"'{PrecomputedValueProvider.__name__}'."
                    )
                    PrecomputedValueProvider.replace_schema_node(node, output_result)
                    # We remove all parent dependencies as the cached output value will
                    # be used.
                    node.needs = {}
                else:
                    # If there is no cached output the node must be re-run if it ends
                    # up as an ancestor of a target node.
                    fingerprint_run_output.is_hit = False

        # Else the node was an input node and the output is the actual node's output.
        else:
            # As fingerprint_run_output is just the node's output there is no need to
            # execute the node again. We can just return it from a
            # `PrecomputedValueProvider`.
            PrecomputedValueProvider.replace_schema_node(node, fingerprint_run_output)
            node.needs = {}

        # Continue walking for every parent node.
        for parent_node_name in node.needs.values():
            self._walk_and_prune(schema, parent_node_name, fingerprint_run_outputs)



from typing import Dict, Any, List

import structlog

from rasa.dialogue_understanding.commands import Command, SetSlotCommand
from rasa.dialogue_understanding.generator import SingleStepLLMCommandGenerator
from rasa.llm_fine_tuning.conversations import Conversation, ConversationStep
from rasa.llm_fine_tuning.paraphrasing.rephrased_user_message import (
    RephrasedUserMessage,
)
from rasa.shared.core.flows import FlowsList
from rasa.shared.exceptions import ProviderClientAPIException
from rasa.shared.utils.llm import (
    llm_factory,
)

structlogger = structlog.get_logger()


class RephraseValidator:
    def __init__(self, llm_config: Dict[str, Any], flows: FlowsList) -> None:
        self.llm_config = llm_config
        self.flows = flows

    async def validate_rephrasings(
        self,
        rephrasings: List[RephrasedUserMessage],
        conversation: Conversation,
    ) -> List[RephrasedUserMessage]:
        """Split rephrased user messages into passing and failing.

        Call an LLM using the same config of the former trained model with an updated
        prompt from the original user message (replace all occurrences of the original
        user message with the rephrased user message). Check if the
        rephrased user message is producing the same commands as the original user
        message. The rephase is passing if the commands match and failing otherwise.

        Args:
            rephrasings: The rephrased user messages.
            conversation: The conversation.

        Returns:
            A list of rephrased user messages including the passing and failing
            rephrases.
        """
        for i, step in enumerate(
            conversation.iterate_over_annotated_user_steps(rephrase=True)
        ):
            current_rephrasings = rephrasings[i]

            for rephrase in current_rephrasings.rephrasings:
                if await self._validate_rephrase_is_passing(rephrase, step):
                    current_rephrasings.passed_rephrasings.append(rephrase)
                else:
                    current_rephrasings.failed_rephrasings.append(rephrase)

        return rephrasings

    async def _validate_rephrase_is_passing(
        self,
        rephrase: str,
        step: ConversationStep,
    ) -> bool:
        prompt = self._update_prompt(
            rephrase, step.original_test_step.text, step.llm_prompt
        )

        action_list = await self._invoke_llm(prompt)

        commands_from_original_utterance = step.llm_commands
        commands_from_rephrased_utterance = (
            SingleStepLLMCommandGenerator.parse_commands(action_list, None, self.flows)
        )
        return self._check_commands_match(
            commands_from_original_utterance, commands_from_rephrased_utterance
        )

    async def _invoke_llm(self, prompt: str) -> str:
        from rasa.dialogue_understanding.generator.constants import DEFAULT_LLM_CONFIG

        llm = llm_factory(self.llm_config, DEFAULT_LLM_CONFIG)

        try:
            llm_response = await llm.acompletion(prompt)
            return llm_response.choices[0]
        except Exception as e:
            # unfortunately, langchain does not wrap LLM exceptions which means
            # we have to catch all exceptions here
            structlogger.error(
                "rephrase_validator.validate_conversation.llm.error", error=e
            )
            raise ProviderClientAPIException(e, message="LLM call exception")

    @staticmethod
    def _check_commands_match(
        expected_commands: List[Command], actual_commands: List[Command]
    ) -> bool:
        if len(expected_commands) != len(actual_commands):
            return False

        for expected_command in expected_commands:
            if isinstance(expected_command, SetSlotCommand):
                slot_name = expected_command.name
                match_found = False
                for c in actual_commands:
                    if isinstance(c, SetSlotCommand) and c.name == slot_name:
                        match_found = True
                        break
                if not match_found:
                    return False

            elif expected_command not in actual_commands:
                return False

        return True

    @staticmethod
    def _update_prompt(
        rephrased_user_message: str, original_user_message: str, prompt: str
    ) -> str:
        return prompt.replace(original_user_message, rephrased_user_message)



from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Text

import structlog

from rasa.dialogue_understanding.commands import (
    Command,
    SetSlotCommand,
    StartFlowCommand,
    ErrorCommand,
)
from rasa.dialogue_understanding.commands.set_slot_command import SetSlotExtractor
from rasa.shared.core.constants import SlotMappingType
from rasa.shared.core.domain import Domain
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.slot_mappings import SlotFillingManager
from rasa.shared.core.trackers import DialogueStateTracker
from rasa.shared.nlu.training_data.message import Message
from rasa.shared.nlu.constants import COMMANDS, TEXT
from rasa.shared.utils.llm import DEFAULT_MAX_USER_INPUT_CHARACTERS
from rasa.shared.constants import (
    RASA_PATTERN_INTERNAL_ERROR_USER_INPUT_TOO_LONG,
    RASA_PATTERN_INTERNAL_ERROR_USER_INPUT_EMPTY,
)

structlogger = structlog.get_logger()


@dataclass
class UserInputConfig:
    """Configuration class for user input settings."""

    max_characters: int = DEFAULT_MAX_USER_INPUT_CHARACTERS
    """The maximum number of characters allowed in the user input."""

    def __post_init__(self) -> None:
        if self.max_characters is None:
            self.max_characters = DEFAULT_MAX_USER_INPUT_CHARACTERS

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "UserInputConfig":
        return cls(
            max_characters=data.get("max_characters", DEFAULT_MAX_USER_INPUT_CHARACTERS)
        )


class CommandGenerator:
    """A command generator.

    Parses a message and returns a list of commands. The commands are then
    executed and will lead to tracker state modifications and action
    predictions.
    """

    def __init__(self, config: Dict[Text, Any]):
        self.user_input_config = UserInputConfig.from_dict(
            config.get("user_input") or {}
        )

    async def process(
        self,
        messages: List[Message],
        flows: FlowsList,
        tracker: Optional[DialogueStateTracker] = None,
        domain: Optional[Domain] = None,
    ) -> List[Message]:
        """Process a list of messages. For each message predict commands.

        The result of the generation is added to the message as a list of
        commands.

        Args:
            messages: The messages to process.
            tracker: The tracker containing the conversation history up to now.
            flows: The flows to use for command prediction.
            domain: The domain.

        Returns:
        The processed messages (usually this is just one during prediction).
        """
        # Determines a set of startable flows by evaluating flow guard conditions.
        startable_flows = self.get_startable_flows(flows, tracker)
        # Get the currently active and called flow (if present).
        # If they would be guarded, e.g. if: false, they would not be in the list
        # of startable flows and not available inside the prompt.
        active_flows = self.get_active_flows(flows, tracker)
        available_flows = FlowsList.from_multiple_flows_lists(
            startable_flows, active_flows
        )

        for message in messages:
            if message.get(COMMANDS):
                # do not overwrite commands if they are already present
                # i.e. another command generator already predicted commands
                continue

            commands = await self._evaluate_and_predict(
                message, available_flows, tracker, domain
            )
            # Double check commands for guarded flows. Unlikely but the llm could
            # have predicted a command for a flow that is not in the startable
            # flow list supplied in the prompt.
            commands = self._check_commands_against_startable_flows(
                commands, startable_flows
            )
            commands = self._check_commands_against_slot_mappings(
                commands, tracker, domain
            )
            commands_dicts = [command.as_dict() for command in commands]
            message.set(COMMANDS, commands_dicts, add_to_output=True)

        return messages

    def get_startable_flows(
        self, flows: FlowsList, tracker: Optional[DialogueStateTracker] = None
    ) -> FlowsList:
        """Determines a set of startable flows by evaluating flow guard conditions.

        Args:
            flows: Underlying flows.
            tracker: The tracker containing the conversation history up to now.

        Returns:
            FlowsList: All flows for which the starting conditions are met.
        """
        if tracker is not None:
            # if tracker is not None, evaluate the flow guard conditions with
            # the current state of the tracker
            return tracker.get_startable_flows(flows)

        # else evaluate it without the tracker context
        return flows.get_startable_flows({})

    def get_active_flows(
        self, flows: FlowsList, tracker: Optional[DialogueStateTracker]
    ) -> FlowsList:
        """Retrieve a list of currently active flows.

        Args:
            flows: Underlying flows.
            tracker: The tracker.

        Returns:
            FlowsList: All currently active flows.
        """
        if not tracker:
            return FlowsList([])
        return tracker.get_active_flows(flows)

    async def _evaluate_and_predict(
        self,
        message: Message,
        startable_flows: FlowsList,
        tracker: Optional[DialogueStateTracker] = None,
        domain: Optional[Domain] = None,
    ) -> List[Command]:
        """Evaluates message for errors and predicts commands if no errors are found.

        Args:
            message: The message to process.
            tracker: The tracker containing the conversation history up to now.
            startable_flows: The startable flows to use for command prediction.

        Returns:
            Errors or predicted commands
        """
        # evaluate message for errors
        if error_commands := self.evaluate_message(message):
            return error_commands

        # if no errors, try predicting commands
        try:
            return await self.predict_commands(
                message, startable_flows, tracker, domain=domain
            )
        except NotImplementedError:
            raise
        except Exception as e:
            structlogger.error("command_generator.predict.error", error=str(e))
            return []

    async def predict_commands(
        self,
        message: Message,
        flows: FlowsList,
        tracker: Optional[DialogueStateTracker] = None,
        **kwargs: Any,
    ) -> List[Command]:
        """Predict commands for a single message.

        Args:
            message: The message to predict commands for.
            flows: The flows to use for command prediction.
            tracker: The tracker containing the conversation history up to now.
            **kwargs: Keyword arguments for forward compatibility.
        Returns:
        The predicted commands.
        """
        raise NotImplementedError()

    def _check_commands_against_startable_flows(
        self, commands: List[Command], startable_flows: FlowsList
    ) -> List[Command]:
        """Check if the start flow commands are only for startable flows.

        Args:
            commands: The commands to check.
            startable_flows: The flows which have their starting conditions satisfied.

        Returns:
            The commands that are startable.
        """
        checked_commands = [
            command
            for command in commands
            if not (
                isinstance(command, StartFlowCommand)
                and command.flow not in startable_flows.flow_ids
            )
        ]

        if commands and len(checked_commands) != len(commands):
            structlogger.info(
                "command_generator.check_commands_against_startable_flows.startable_commands",
                commands=checked_commands,
            )

        return checked_commands

    def evaluate_message(self, message: Message) -> List[Command]:
        """Evaluates the given message.

        Args:
            message: The message to evaluate.

        Returns:
            A list of error commands indicating the type of error.
        """
        errors: List[Command]

        if self.check_if_message_is_empty(message):
            # notify the user that the message is empty
            errors = [
                ErrorCommand(error_type=RASA_PATTERN_INTERNAL_ERROR_USER_INPUT_EMPTY)
            ]
        elif self.check_if_message_exceeds_limit(message):
            # notify the user about message length
            errors = [
                ErrorCommand(
                    error_type=RASA_PATTERN_INTERNAL_ERROR_USER_INPUT_TOO_LONG,
                    info={"max_characters": self.user_input_config.max_characters},
                )
            ]
        else:
            return []

        structlogger.info(
            "command_generator.evaluate_message.error",
            event_info="Invalid message",
            errors=[e.as_dict() for e in errors],
            message=message.get(TEXT),
        )
        return errors

    def check_if_message_exceeds_limit(self, message: Message) -> bool:
        """Checks if the given message exceeds the predefined number of characters."""
        # if limit was a negative number, omit it
        if self.user_input_config.max_characters < 0:
            return False
        return len(message.get(TEXT, "")) > self.user_input_config.max_characters

    def check_if_message_is_empty(self, message: Message) -> bool:
        """Checks if the given message is empty or whitespace-only."""
        return len(message.get(TEXT, "").strip()) == 0

    @staticmethod
    def _check_commands_against_slot_mappings(
        commands: List[Command],
        tracker: DialogueStateTracker,
        domain: Optional[Domain] = None,
    ) -> List[Command]:
        """Check if the LLM-issued slot commands are fillable.

        The LLM-issued slot commands are fillable if the slot
        mappings are satisfied.
        """
        if not domain:
            return commands

        llm_fillable_slot_names = [
            command.name
            for command in commands
            if isinstance(command, SetSlotCommand)
            and command.extractor == SetSlotExtractor.LLM.value
        ]

        if not llm_fillable_slot_names:
            return commands

        llm_fillable_slots = [
            slot for slot in domain.slots if slot.name in llm_fillable_slot_names
        ]

        slot_filling_manager = SlotFillingManager(domain, tracker)
        slots_to_be_removed = []

        structlogger.debug(
            "command_processor.check_commands_against_slot_mappings.active_flow",
            active_flow=tracker.active_flow,
        )

        for slot in llm_fillable_slots:
            should_fill_slot = False
            for mapping in slot.mappings:
                mapping_type = SlotMappingType(mapping.get("type"))

                should_fill_slot = slot_filling_manager.should_fill_slot(
                    slot.name, mapping_type, mapping
                )

                if should_fill_slot:
                    break

            if not should_fill_slot:
                structlogger.debug(
                    "command_processor.check_commands_against_slot_mappings.slot_not_fillable",
                    slot_name=slot.name,
                )
                slots_to_be_removed.append(slot.name)

        if not slots_to_be_removed:
            return commands

        filtered_commands = [
            command
            for command in commands
            if not (
                isinstance(command, SetSlotCommand)
                and command.name in slots_to_be_removed
            )
        ]

        return filtered_commands



import sys
from pathlib import Path
from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union

import structlog

import rasa.shared.utils.io
from rasa.e2e_test.constants import SCHEMA_FILE_PATH
from rasa.shared.utils.yaml import read_schema_file

if TYPE_CHECKING:
    from rasa.e2e_test.e2e_test_case import TestCase


structlogger = structlog.get_logger()


def validate_path_to_test_cases(path: str) -> None:
    """Validate that path to test cases exists."""
    if not Path(path).exists():
        rasa.shared.utils.io.raise_warning(
            f"Path to test cases does not exist: {path}. "
            f"Please provide a valid path to test cases. "
            f"Exiting...",
            UserWarning,
        )
        sys.exit(1)


def validate_test_case(test_case_name: str, input_test_cases: List["TestCase"]) -> None:
    """Validate that test case exists."""
    if test_case_name and not input_test_cases:
        rasa.shared.utils.io.raise_warning(
            f"Test case does not exist: {test_case_name}. "
            f"Please check for typos and provide a valid test case name. "
            f"Exiting...",
            UserWarning,
        )
        sys.exit(1)


def validate_model_path(model_path: Optional[str], parameter: str, default: str) -> str:
    """Validate the model path.

    Args:
        model_path: Path to the model.
        parameter: Name of the parameter.
        default: Default path to the model.

    Returns:
    Path to the model.
    """
    if model_path and Path(model_path).exists():
        return model_path

    if model_path and not Path(model_path).exists():
        rasa.shared.utils.io.raise_warning(
            f"The provided model path '{model_path}' could not be found. "
            f"Using default location '{default}' instead.",
            UserWarning,
        )

    elif model_path is None:
        structlogger.info(
            "rasa.e2e_test.validate_model_path",
            message=f"Parameter '{parameter}' is not set. "
            f"Using default location '{default}' instead.",
        )

    Path(default).mkdir(exist_ok=True)
    return default


def read_e2e_test_schema() -> Union[List[Any], Dict[str, Any]]:
    """Read the schema for the e2e test files.

    Returns:
        The content of the schema.
    """
    return read_schema_file(SCHEMA_FILE_PATH)



from pathlib import Path
from typing import Dict, List, Optional, Text, Union

import rasa.shared.data
import rasa.shared.utils.io
from rasa.shared.core.domain import (
    KEY_ACTIONS,
    KEY_ENTITIES,
    KEY_FORMS,
    KEY_INTENTS,
    KEY_RESPONSES,
    KEY_SLOTS,
    Domain,
)
from rasa.shared.core.flows.yaml_flows_io import KEY_FLOWS, is_flows_file
from rasa.shared.core.training_data.story_reader.yaml_story_reader import (
    KEY_RULE_NAME,
    KEY_RULES,
    KEY_STORIES,
    KEY_STORY_NAME,
    YAMLStoryReader,
)
from rasa.shared.nlu.training_data.formats.rasa_yaml import (
    KEY_INTENT,
    KEY_NLU,
)
from rasa.shared.utils.yaml import read_yaml_file


class RasaPrimitiveStorageMapper:
    """Maps the Rasa primitives to the file its located in."""

    def __init__(
        self,
        domain_path: Optional[Union[Text, Path]] = None,
        training_data_paths: Optional[Union[List[Text], List[Path], Text, Path]] = None,
    ):
        self._primitives: Dict[str, Dict] = {
            "entities": {},
            "slots": {},
            "forms": {},
            "intents": {},
            "stories": {},
            "rules": {},
            "actions": {},
            "responses": {},
            "flows": {},
        }

        if domain_path:
            self._load_domain(domain_path)

        if training_data_paths:
            self._load_training_data(training_data_paths)

    def _load_domain(self, domain_path: Union[Text, Path]) -> None:
        self._resolve_domain_files(domain_path)
        self._parse_domain_files()

    def _load_training_data(
        self, training_data_paths: Union[List[Text], List[Path], Text, Path]
    ) -> None:
        self._resolve_training_data_files(training_data_paths)

        self._nlu_files = rasa.shared.data.get_data_files(
            self._training_data_paths, rasa.shared.data.is_nlu_file
        )
        self._parse_nlu_data()

        self._story_files = rasa.shared.data.get_data_files(
            self._training_data_paths, YAMLStoryReader.is_stories_file
        )
        self._parse_story_data()

        self._flow_files = rasa.shared.data.get_data_files(
            self._training_data_paths, is_flows_file
        )
        self._parse_flow_data()

    def _resolve_training_data_files(
        self, training_data_paths: Union[List[Text], List[Path], Text, Path]
    ) -> None:
        if isinstance(training_data_paths, list):
            training_data_paths = [Path(p) for p in training_data_paths]
        elif isinstance(training_data_paths, str):
            training_data_paths = [Path(training_data_paths)]
        elif isinstance(training_data_paths, Path):
            training_data_paths = [training_data_paths]

        self._training_data_paths = training_data_paths

    def _resolve_domain_files(self, domain_path: Union[Text, Path]) -> None:
        domain_path = Path(domain_path) if isinstance(domain_path, str) else domain_path

        if domain_path.is_dir():
            self._domain_files = [
                file for file in domain_path.iterdir() if Domain.is_domain_file(file)
            ]
        else:
            self._domain_files = [domain_path]

    def _parse_domain_files(self) -> None:
        for domain_file in self._domain_files:
            content = read_yaml_file(domain_file)
            for entity in content.get(KEY_ENTITIES, []):  # type: ignore[union-attr]
                self._assign_value(
                    self._primitives["entities"], "domain", domain_file, entity
                )
            for intent in content.get(KEY_INTENTS, []):  # type: ignore[union-attr]
                self._assign_value(
                    self._primitives["intents"], "domain", domain_file, intent
                )
            for slot in content.get(KEY_SLOTS, []):  # type: ignore[union-attr]
                self._assign_value(
                    self._primitives["slots"], "domain", domain_file, slot
                )
            for form in content.get(KEY_FORMS, []):  # type: ignore[union-attr]
                self._assign_value(
                    self._primitives["forms"], "domain", domain_file, form
                )
            for action in content.get(KEY_ACTIONS, []):  # type: ignore[union-attr]
                self._assign_value(
                    self._primitives["actions"], "domain", domain_file, action
                )
            for response in content.get(KEY_RESPONSES, []):  # type: ignore[union-attr]
                self._assign_value(
                    self._primitives["responses"], "domain", domain_file, response
                )

    def _parse_nlu_data(self) -> None:
        """Parses the nlu data and extracts the intents."""
        for nlu_file in self._nlu_files:
            content = read_yaml_file(nlu_file)
            for intent in content.get(KEY_NLU, []):  # type: ignore[union-attr]
                if KEY_INTENT in intent:
                    self._assign_value(
                        self._primitives["intents"],
                        "training",
                        nlu_file,
                        intent.get(KEY_INTENT),
                    )

    def _parse_story_data(self) -> None:
        """Parses the story data and extracts the stories and rules."""
        for story_file in self._story_files:
            content = read_yaml_file(story_file)
            for story in content.get(KEY_STORIES, []):  # type: ignore[union-attr]
                self._assign_value(
                    self._primitives["stories"],
                    "training",
                    story_file,
                    story[KEY_STORY_NAME],
                )
            for rule in content.get(KEY_RULES, []):  # type: ignore[union-attr]
                self._assign_value(
                    self._primitives["rules"],
                    "training",
                    story_file,
                    rule[KEY_RULE_NAME],
                )

    def _parse_flow_data(self) -> None:
        """Parses the flow data and extracts the flow ids."""
        for flow_file in self._flow_files:
            content = read_yaml_file(flow_file)
            for flow in content.get(KEY_FLOWS, []):  # type: ignore[union-attr]
                self._assign_value(
                    self._primitives["flows"],
                    "training",
                    flow_file,
                    flow,
                )

    @staticmethod
    def _assign_value(
        primitive: Dict, file_type: str, value: Union[str, Path], key: str
    ) -> None:
        if not isinstance(value, Path):
            value = Path(value)

        if isinstance(key, dict):
            key = next(iter(key.keys()))

        if key in primitive:
            if file_type in primitive[key]:
                primitive[key][file_type].append(value)
            else:
                primitive[key][file_type] = [value]
        else:
            primitive[key] = {file_type: [value]}

    def get_file(self, primitive: Text, primitive_type: Text) -> Dict[str, List[Path]]:
        """Returns the file where the primitive is located.

        If the primitive is not found, empty dict is returned.

        Args:
            primitive: The name or ID of the primitive to search for.
            primitive_type: The type of the primitive to search for.
                either entities, slots, forms, intents, stories or rules.

        Returns:
            A dictionary containing the file type and the list of file paths.
            file_type can be either domain or training.
            get_file()[file_type] -> List[Path]
        """
        try:
            return self._primitives[primitive_type][primitive]
        except KeyError:
            return {}



import base64
import logging
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import rasa.shared.utils.cli
import requests
from rasa.shared.core.domain import KEY_RESPONSES, KEY_SLOTS, Domain
from rasa.shared.core.flows.flow import Flow
from rasa.shared.exceptions import RasaException
from rasa.shared.importers.importer import TrainingDataImporter
from rasa.shared.utils.yaml import read_yaml, write_yaml
from rasa.utils.common import get_temp_dir_name

from rasa.studio.auth import KeycloakToken, KeycloakTokenReader, StudioAuth
from rasa.studio.config import StudioConfig

logger = logging.getLogger(__name__)


class StudioDataHandler:
    def __init__(
        self, studio_config: StudioConfig, assistant_name: Optional[str]
    ) -> None:
        self.studio_config = studio_config.read_config()
        if not assistant_name:
            rasa.shared.utils.cli.print_error_and_exit(
                "Assistant name not specified please specify"
                "the assistant name with: 'rasa studio <command> assistant_name'"
            )
        self.assistant_name = assistant_name

        self.nlu: Optional[str] = None
        self.domain: Optional[str] = None
        self.flows: Optional[str] = None

    def has_flows(self) -> bool:
        return self.flows is not None

    def has_nlu(self) -> bool:
        return self.nlu is not None

    def _build_request(
        self,
        intent_names: Optional[List[str]] = None,
        entity_names: Optional[List[str]] = None,
    ) -> dict:
        request = {
            "query": (
                "query ExportAsEncodedYaml($input: ExportAsEncodedYamlInput!) "
                "{ exportAsEncodedYaml(input: $input) "
                "{ ... on ExportModernAsEncodedYamlOutput "
                "{ nlu flows domain endpoints config } "
                "... on ExportClassicAsEncodedYamlOutput "
                "{ nlu domain }}}"
            ),
            "variables": {"input": {"assistantName": self.assistant_name}},
        }
        if intent_names or entity_names:
            obj = []
            if intent_names:
                obj.append(
                    {
                        "name": intent_names,
                        "type": "Intent",
                    }
                )
            if entity_names:
                obj.append(
                    {
                        "name": entity_names,
                        "type": "Entity",
                    }
                )
            request["variables"]["input"]["objects"] = obj  # type: ignore[index]

        return request

    def _make_request(
        self, GQL_req: Dict[Any, Any], verify: bool = True
    ) -> Dict[Any, Any]:
        token = KeycloakTokenReader().get_token()
        if token.is_expired():
            token = self.refresh_token(token)

        if not self.studio_config.studio_url:
            return rasa.shared.utils.cli.print_error_and_exit(
                "No endpoint found in config. Please run `rasa studio config`."
            )

        res = requests.post(
            url=self.studio_config.studio_url,
            json=GQL_req,
            headers={
                "Authorization": f"{token.token_type} {token.access_token}",
                "Content-Type": "application/json",
            },
            verify=verify,
        )

        if res.status_code != 200:
            raise RasaException(
                f"Download from Studio with URL: "
                f"{self.studio_config.studio_url} failed "
                f"with status code {res.status_code}"
            )

        res_json = res.json()
        if self._validate_response(res_json):
            return res_json

        return rasa.shared.utils.cli.print_error_and_exit(
            "Failed to download data from Rasa Studio."
        )

    def refresh_token(self, token: KeycloakToken) -> KeycloakToken:
        if not token.can_refresh():
            rasa.shared.utils.cli.print_error_and_exit(
                "Access token expired and cannot be refreshed. "
                "Please run `rasa studio login`."
            )
        auth = StudioAuth(self.studio_config)
        auth.refresh_token(token.refresh_token)
        token = KeycloakTokenReader().get_token()
        return token

    def request_all_data(self) -> None:
        """Gets the data from Rasa Studio.

        Returns:
            The data from Rasa Studio.
        """
        GQL_req = self._build_request()
        verify = not self.studio_config.disable_verify

        response = self._make_request(GQL_req, verify=verify)
        self._extract_data(response)

    def request_data(
        self, intent_names: List[str], entity_names: List[str], **kwargs: Any
    ) -> None:
        """Gets the data from Rasa Studio.

        Args:
            intent_names: List of intents to download
            entity_names: List of entities to download
            **kwargs: Additional arguments to pass to the request

        Returns:
            The data from Rasa Studio.
        """
        GQL_req = self._build_request(intent_names, entity_names)
        verify = not self.studio_config.disable_verify

        response = self._make_request(GQL_req, verify=verify)
        self._extract_data(response)

    def get_config(self) -> Optional[str]:
        return self.config

    def get_endpoints(self) -> Optional[str]:
        return self.endpoints

    def _validate_response(self, response: dict) -> bool:
        """Validates the response from Rasa Studio.

        Args:
            response: The response from Rasa Studio.

        Returns:
            True if the response is valid, False otherwise.
        """
        try:
            if not response.get("data"):
                for error in response["errors"]:
                    logger.error(error["message"])
                return False
            if not response["data"]["exportAsEncodedYaml"]:
                logger.error(response["data"])
                return False
            if not (
                response["data"]["exportAsEncodedYaml"].get("nlu")
                or response["data"]["exportAsEncodedYaml"].get("flows")
            ):
                logger.error(
                    "No nlu or flows data in Studio response."
                    "Cannot determine assistant type."
                )
                return False
        except KeyError:
            return False
        return True

    def _extract_data(self, response: dict) -> None:
        return_data = response["data"]["exportAsEncodedYaml"]

        self.nlu = self._decode_response(return_data.get("nlu"))
        self.domain = self._decode_response(return_data.get("domain"))
        self.flows = self._decode_response(return_data.get("flows"))
        self.config = self._decode_response(return_data.get("config"))
        self.endpoints = self._decode_response(return_data.get("endpoints"))

        if not self.has_nlu() and not self.has_flows():
            raise RasaException("No nlu or flows data in Studio response.")

    @staticmethod
    def _decode_response(data: str) -> Optional[str]:
        if not data:
            return None
        return base64.b64decode(data).decode("utf-8")


def combine_domains(
    studio_domain: Dict[str, Any], original_domain: Dict[str, Any]
) -> Dict:
    """Create a new domain file from the diff."""
    if studio_domain is None or original_domain is None:
        return {}
    return _combine_domain_keys(studio_domain, original_domain)


def _combine_domain_keys(
    first_domain: Dict[str, Any], second_domain: Dict[str, Any]
) -> Dict[str, Any]:
    combined_keys = {}
    for key in first_domain:
        if key not in second_domain:
            combined_keys[key] = first_domain[key]
        elif isinstance(first_domain[key], dict):
            combined_keys[key] = _combine_domain_keys(
                first_domain[key], second_domain[key]
            )
            # remove empty diffs
            if not combined_keys[key]:
                del combined_keys[key]
            elif key not in [KEY_SLOTS, KEY_RESPONSES]:
                # for all keys except slots and responses, we want to keep the
                # keys from the first domain
                combined_keys[key] = first_domain[key]
        elif isinstance(first_domain[key], list):
            combined_keys[key] = []
            for item in first_domain[key]:
                if item not in second_domain[key]:
                    combined_keys[key].append(item)

            # if list is empty, remove it
            if not combined_keys[key]:
                del combined_keys[key]

    return combined_keys


def _diff_nlu_examples(
    new_example: Dict,
    nlu_diff: List,
    match_key: str,
    match_value: str,
    original_nlu_examples: List,
) -> None:
    """Creates diff of nlu data examples.

    Args:
        new_example (Dict): intent or synonym with examples
        nlu_diff (List): list of diff examples
        match_key (str): intent or synonym name
        match_value (str): intent or synonym value
        original_nlu_examples (List): original nlu examples
    """
    orig = list(
        filter(
            lambda x: x.get(match_key) == match_value,
            original_nlu_examples,
        )
    )
    if len(orig) == 1:
        orig_ex = orig[0]["examples"].split("\n")
        new_ex = new_example["examples"].split("\n")
        new_example["examples"] = "\n".join(list(set(new_ex).difference(set(orig_ex))))
        if not new_example["examples"]:
            nlu_diff.remove(new_example)


def create_new_nlu_from_diff(
    studio_nlu: Dict[str, Any], original_nlu: Dict[str, Any]
) -> Dict:
    """Create a new nlu file from the diff."""
    # `or []` handles the case where the data contains the property as an empty
    # key, example yaml:
    # ```
    # nlu:
    # ```
    # in this case, the yaml parser will return an empty dict (because it
    # can't know that it is supposed to be a list, so we need to convert it
    # to a list
    studio_nlu_data = studio_nlu.get("nlu", []) or []
    original_nlu_data = original_nlu.get("nlu", []) or []

    nlu_diff = []
    for new in studio_nlu_data:
        if new in original_nlu_data:
            continue

        nlu_diff.append(new)
        intent = new.get("intent")
        if intent:
            _diff_nlu_examples(new, nlu_diff, "intent", intent, original_nlu_data)
        synonym = new.get("synonym")
        if synonym:
            _diff_nlu_examples(new, nlu_diff, "synonym", synonym, original_nlu_data)

    return {"nlu": nlu_diff}


def create_new_flows_from_diff(
    studio_flows: List[Flow], original_flows: List[Flow]
) -> List[Flow]:
    """Create a new flows file from the diff."""
    flows_new = [new for new in studio_flows if new not in original_flows]
    return flows_new


def import_data_from_studio(
    handler: StudioDataHandler, domain_path: Path, data_paths: List[Path]
) -> Tuple[TrainingDataImporter, TrainingDataImporter]:
    """Construct TrainingDataImporter from Studio data and original data.

    Args:
        handler (StudioDataHandler): handler with data from studio
        domain_path (Path): Path to a domain file
        data_paths (List[Path]): List of paths to training data files

    Returns:
        Tuple[TrainingDataImporter, TrainingDataImporter]:
        data from studio and original data
    """
    tmp_dir = get_temp_dir_name()
    data_original = TrainingDataImporter.load_from_dict(
        domain_path=domain_path, training_data_paths=data_paths
    )

    data_paths = []

    if handler.has_nlu():
        nlu_file = Path(tmp_dir, "nlu.yml")
        write_yaml(read_yaml(handler.nlu), nlu_file)
        data_paths.append(str(nlu_file))

    if handler.has_flows():
        flows_file = Path(tmp_dir, "flows.yml")
        write_yaml(read_yaml(handler.flows), flows_file)
        data_paths.append(str(flows_file))

    if not data_paths:
        return rasa.shared.utils.cli.print_error_and_exit(
            "No Data for nlu or flows. Can't import data from Studio."
        )

    studio_domain = Domain.from_yaml(handler.domain)
    domain_file = Path(tmp_dir, "domain.yml")
    studio_domain.persist(domain_file)

    data_from_studio = TrainingDataImporter.load_from_dict(
        domain_path=str(domain_file), training_data_paths=data_paths
    )

    return data_from_studio, data_original



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List

import structlog
from rasa.dialogue_understanding.commands import Command
from rasa.dialogue_understanding.patterns.clarify import ClarifyPatternFlowStackFrame
from rasa.shared.core.events import Event
from rasa.shared.core.flows import FlowsList
from rasa.shared.core.trackers import DialogueStateTracker

structlogger = structlog.get_logger()


@dataclass
class ClarifyCommand(Command):
    """A command to indicate that the bot should ask for clarification."""

    options: List[str]

    @classmethod
    def command(cls) -> str:
        """Returns the command type."""
        return "clarify"

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> ClarifyCommand:
        """Converts the dictionary to a command.

        Returns:
            The converted dictionary.
        """
        try:
            return ClarifyCommand(options=data["options"])
        except KeyError as e:
            raise ValueError(
                f"Missing parameter '{e}' while parsing ClarifyCommand."
            ) from e

    def run_command_on_tracker(
        self,
        tracker: DialogueStateTracker,
        all_flows: FlowsList,
        original_tracker: DialogueStateTracker,
    ) -> List[Event]:
        """Runs the command on the tracker.

        Args:
            tracker: The tracker to run the command on.
            all_flows: All flows in the assistant.
            original_tracker: The tracker before any command was executed.

        Returns:
            The events to apply to the tracker.
        """

        flows = [all_flows.flow_by_id(opt) for opt in self.options]
        clean_options = [flow.id for flow in flows if flow is not None]
        if len(clean_options) != len(self.options):
            structlogger.debug(
                "command_executor.altered_command.dropped_clarification_options",
                command=self,
                original_options=self.options,
                cleaned_options=clean_options,
            )
        if len(clean_options) == 0:
            structlogger.debug(
                "command_executor.skip_command.empty_clarification", command=self
            )
            return []

        stack = tracker.stack
        relevant_flows = [all_flows.flow_by_id(opt) for opt in clean_options]
        names = [flow.readable_name() for flow in relevant_flows if flow is not None]
        stack.push(ClarifyPatternFlowStackFrame(names=names))
        return tracker.create_stack_updated_events(stack)

    def __hash__(self) -> int:
        return hash(tuple(self.options))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ClarifyCommand):
            return False

        return other.options == self.options



from __future__ import annotations

import copy
from dataclasses import dataclass, field
from functools import cached_property
from pathlib import Path
from typing import Text, Optional, Dict, Any, List, Set, Union

import structlog
from pypred import Predicate

import rasa.shared.utils.io
from rasa.shared.constants import RASA_DEFAULT_FLOW_PATTERN_PREFIX
from rasa.shared.core.flows.flow_path import PathNode, FlowPath, FlowPathsList
from rasa.shared.core.flows.flow_step import FlowStep
from rasa.shared.core.flows.flow_step_links import (
    FlowStepLink,
    StaticFlowStepLink,
    IfFlowStepLink,
    ElseFlowStepLink,
)
from rasa.shared.core.flows.flow_step_sequence import FlowStepSequence
from rasa.shared.core.flows.nlu_trigger import NLUTriggers
from rasa.shared.core.flows.steps import (
    CollectInformationFlowStep,
    EndFlowStep,
    StartFlowStep,
    ActionFlowStep,
)
from rasa.shared.core.flows.steps.constants import (
    CONTINUE_STEP_PREFIX,
    START_STEP,
    END_STEP,
)
from rasa.shared.core.flows.steps.continuation import ContinueFlowStep
from rasa.shared.core.slots import Slot

structlogger = structlog.get_logger()


@dataclass
class Flow:
    """Represents the configuration of a flow."""

    id: Text
    """The id of the flow."""
    custom_name: Optional[Text] = None
    """The human-readable name of the flow."""
    description: Optional[Text] = None
    """The description of the flow."""
    guard_condition: Optional[Text] = None
    """The condition that needs to be fulfilled for the flow to be startable."""
    step_sequence: FlowStepSequence = field(default_factory=FlowStepSequence.empty)
    """The steps of the flow."""
    nlu_triggers: Optional[NLUTriggers] = None
    """The list of intents, e.g. nlu triggers, that start the flow."""
    always_include_in_prompt: Optional[bool] = None
    """
    A flag that checks whether the flow should always be included in the prompt or not.
    """
    file_path: Optional[str] = None
    """The path to the file where the flow is stored."""

    @staticmethod
    def from_json(
        flow_id: Text,
        data: Dict[Text, Any],
        file_path: Optional[Union[str, Path]] = None,
    ) -> Flow:
        """Create a Flow object from serialized data.

        Args:
            flow_id: id of the flow
            data: data for a Flow object in a serialized format.
            file_path: the file path of the flow

        Returns:
            A Flow object.
        """
        step_sequence = FlowStepSequence.from_json(flow_id, data.get("steps"))
        nlu_triggers = NLUTriggers.from_json(data.get("nlu_trigger"))

        if file_path and isinstance(file_path, Path):
            file_path = str(file_path)

        return Flow(
            id=flow_id,
            custom_name=data.get("name"),
            description=data.get("description"),
            always_include_in_prompt=data.get("always_include_in_prompt"),
            # str or bool are permitted in the flow schema but internally we want a str
            guard_condition=str(data["if"]) if "if" in data else None,
            step_sequence=Flow.resolve_default_ids(step_sequence),
            nlu_triggers=nlu_triggers,
            # If we are reading the flows in after training the file_path is part of
            # data. When the model is trained, take the provided file_path.
            file_path=data.get("file_path") if "file_path" in data else file_path,
        )

    def get_full_name(self) -> str:
        if self.file_path:
            return f"{self.file_path}::{self.name}"
        return self.name

    @staticmethod
    def create_default_name(flow_id: str) -> str:
        """Create a default flow name for when it is missing."""
        return flow_id.replace("_", " ").replace("-", " ")

    @staticmethod
    def resolve_default_ids(step_sequence: FlowStepSequence) -> FlowStepSequence:
        """Resolves the default ids of all steps in the sequence.

        If a step does not have an id, a default id is assigned to it based
        on the type of the step and its position in the flow.

        Similarly, if a step doesn't have an explicit next assigned we resolve
        the default next step id.

        Args:
            step_sequence: The step sequence to resolve the default ids for.

        Returns:
            The step sequence with the default ids resolved.
        """
        # assign an index to all steps
        for idx, step in enumerate(step_sequence.steps):
            step.idx = idx

        def resolve_default_next(steps: List[FlowStep], is_root_sequence: bool) -> None:
            for i, step in enumerate(steps):
                if step.next.no_link_available() and step.does_allow_for_next_step():
                    if i == len(steps) - 1:
                        if is_root_sequence:
                            # if this is the root sequence, we need to add an end step
                            # to the end of the sequence. other sequences, e.g.
                            # in branches need to explicitly add a next step.
                            step.next.links.append(StaticFlowStepLink(END_STEP))
                    else:
                        step.next.links.append(StaticFlowStepLink(steps[i + 1].id))
                for link in step.next.links:
                    if sub_steps := link.child_steps():
                        resolve_default_next(sub_steps, is_root_sequence=False)

        resolve_default_next(step_sequence.child_steps, is_root_sequence=True)
        return step_sequence

    def as_json(self) -> Dict[Text, Any]:
        """Serialize the Flow object.

        Returns:
            The Flow object as serialized data.
        """
        data: Dict[Text, Any] = {
            "id": self.id,
            "steps": self.step_sequence.as_json(),
        }
        if self.custom_name is not None:
            data["name"] = self.custom_name
        if self.description is not None:
            data["description"] = self.description
        if self.guard_condition is not None:
            data["if"] = self.guard_condition
        if self.always_include_in_prompt is not None:
            data["always_include_in_prompt"] = self.always_include_in_prompt
        if self.nlu_triggers:
            data["nlu_trigger"] = self.nlu_triggers.as_json()
        if self.file_path:
            data["file_path"] = self.file_path

        return data

    def readable_name(self) -> str:
        """Returns the name of the flow or its id if no name is set."""
        return self.name or self.id

    def step_by_id(self, step_id: Optional[Text]) -> Optional[FlowStep]:
        """Returns the step with the given id."""
        if not step_id:
            return None

        if step_id == START_STEP:
            return StartFlowStep(self.id, self.first_step_in_flow().id)

        if step_id == END_STEP:
            return EndFlowStep(self.id)

        if step_id.startswith(CONTINUE_STEP_PREFIX):
            return ContinueFlowStep(self.id, step_id[len(CONTINUE_STEP_PREFIX) :])

        for step in self.steps_with_calls_resolved:
            if step.id == step_id:
                return step

        return None

    def first_step_in_flow(self) -> FlowStep:
        """Returns the start step of this flow."""
        if not (steps := self.steps):
            raise RuntimeError(
                f"Flow {self.id} is empty despite validation that this cannot happen."
            )
        return steps[0]

    def get_trigger_intents(self) -> Set[str]:
        """Returns the trigger intents of the flow."""
        results: Set[str] = set()

        if not self.nlu_triggers:
            return results

        for condition in self.nlu_triggers.trigger_conditions:
            results.add(condition.intent)

        return results

    @property
    def is_rasa_default_flow(self) -> bool:
        """Test whether the flow is a rasa default flow."""
        return self.id.startswith(RASA_DEFAULT_FLOW_PATTERN_PREFIX)

    def get_collect_steps(self) -> List[CollectInformationFlowStep]:
        """Return all CollectInformationFlowSteps in the flow."""
        collect_steps = []
        for step in self.steps_with_calls_resolved:
            if isinstance(step, CollectInformationFlowStep):
                collect_steps.append(step)
        return collect_steps

    @property
    def steps_with_calls_resolved(self) -> List[FlowStep]:
        """Return the steps of the flow including steps of called flows."""
        return self.step_sequence.steps_with_calls_resolved

    @property
    def steps(self) -> List[FlowStep]:
        """Return the steps of the flow without steps of called flows."""
        return self.step_sequence.steps

    @cached_property
    def fingerprint(self) -> str:
        """Create a fingerprint identifying this step sequence."""
        return rasa.shared.utils.io.deep_container_fingerprint(self.as_json())

    @property
    def utterances(self) -> Set[str]:
        """Retrieve all utterances of this flow."""
        return set().union(
            *[step.utterances for step in self.step_sequence.steps_with_calls_resolved]
        )

    @property
    def custom_actions(self) -> Set[str]:
        """Retrieve all custom actions of this flow."""
        return {
            step.custom_action
            for step in self.step_sequence.steps_with_calls_resolved
            if isinstance(step, ActionFlowStep) and step.custom_action is not None
        }

    @property
    def name(self) -> str:
        """Create a default name if none is present."""
        return self.custom_name or Flow.create_default_name(self.id)

    def is_startable(
        self,
        context: Optional[Dict[Text, Any]] = None,
        slots: Optional[Dict[Text, Slot]] = None,
    ) -> bool:
        """Return whether the start condition is satisfied.

        Args:
            context: The context data to evaluate the starting conditions against.
            slots: The slots to evaluate the starting conditions against.

        Returns:
            Whether the start condition is satisfied.
        """
        context = context or {}
        slots = slots or {}
        simplified_slots = {slot.name: slot.value for slot in slots.values()}

        # If no starting condition exists, the flow is always startable.
        if not self.guard_condition:
            return True

        # if a flow guard condition exists and the flow was started via a link,
        # e.g. is currently active, the flow is startable
        if context.get("flow_id") == self.id:
            return True

        try:
            predicate = Predicate(self.guard_condition)
            is_startable = predicate.evaluate(
                {"context": context, "slots": simplified_slots}
            )
            structlogger.debug(
                "command_generator.validate_flow_starting_conditions.result",
                predicate=predicate.description(),
                is_startable=is_startable,
                flow_id=self.id,
            )
            return is_startable
        # if there is any kind of exception when evaluating the predicate, the flow
        # is not startable
        except (TypeError, Exception) as e:
            structlogger.error(
                "command_generator.validate_flow_starting_conditions.error",
                predicate=self.guard_condition,
                context=context,
                slots=slots,
                error=str(e),
            )
            return False

    def has_action_step(self, action: Text) -> bool:
        """Check whether the flow has an action step with the given action."""
        for step in self.steps:
            if isinstance(step, ActionFlowStep) and step.action == action:
                return True
        return False

    def is_startable_only_via_link(self) -> bool:
        """Determines if the flow can be initiated exclusively through a link.

        This condition is met when a guard condition exists and is
        consistently evaluated to `False` (e.g. `if: False`).

        Returns:
            A boolean indicating if the flow initiation is link-based only.
        """
        if (
            self.guard_condition is None
            or self._contains_variables_in_guard_condition()
        ):
            return False

        try:
            predicate = Predicate(self.guard_condition)
            is_startable_via_link = not predicate.evaluate({})
            structlogger.debug(
                "flow.is_startable_only_via_link.result",
                predicate=self.guard_condition,
                is_startable_via_link=is_startable_via_link,
                flow_id=self.id,
            )
            return is_startable_via_link
        # if there is any kind of exception when evaluating the predicate, the flow
        # is not startable by link or by any other means.
        except (TypeError, Exception) as e:
            structlogger.error(
                "flow.is_startable_only_via_link.error",
                predicate=self.guard_condition,
                error=str(e),
                flow_id=self.id,
            )
            return False

    def _contains_variables_in_guard_condition(self) -> bool:
        """Determines if the guard condition contains dynamic literals.

        I.e. literals that cannot be statically resolved, indicating a variable.

        Returns:
            True if dynamic literals are present, False otherwise.
        """
        from pypred import ast
        from pypred.tiler import tile, SimplePattern

        if not self.guard_condition:
            return False

        predicate = Predicate(self.guard_condition)

        # find all literals in the AST tree
        literals = []
        tile(
            ast=predicate.ast,
            patterns=[SimplePattern("types:Literal")],
            func=lambda _, literal: literals.append(literal),
        )

        # check if there is a literal that cannot be statically resolved (variable)
        for literal in literals:
            if type(predicate.static_resolve(literal.value)) == ast.Undefined:
                return True

        return False

    def extract_all_paths(self) -> FlowPathsList:
        """Extracts all possible flow paths.

        Extracts all possible flow paths from a given flow structure by
        recursively exploring each step.
        This function initializes an empty list to collect paths, an empty path list,
        and a set of visited step IDs to prevent revisiting steps.
        It calls `go_over_steps` to recursively explore and fill the paths list.
        """
        flow_paths_list = FlowPathsList(self.id, paths=[])
        steps: List[FlowStep] = self.steps
        current_path: FlowPath = FlowPath(flow=self.id, nodes=[])
        step_ids_visited: Set[str] = set()

        self._go_over_steps(steps, current_path, flow_paths_list, step_ids_visited)

        if not flow_paths_list.is_path_part_of_list(current_path):
            flow_paths_list.paths.append(copy.deepcopy(current_path))

        structlogger.debug(
            "shared.core.flows.flow.extract_all_paths",
            comment="Extraction complete",
            number_of_paths=len(flow_paths_list.paths),
            flow_name=self.name,
        )
        return flow_paths_list

    def _go_over_steps(
        self,
        steps_to_go: Union[str, List[FlowStep]],
        current_path: FlowPath,
        completed_paths: FlowPathsList,
        step_ids_visited: Set[str],
    ) -> None:
        """Processes the flow steps recursively.

        Either following direct step IDs or handling conditions, and adds complete
        paths to the collected_paths.

        Args:
            steps_to_go: Either a direct step ID or a list of steps to process.
            current_path: The current path being constructed.
            completed_paths: The list where completed paths are added.
            step_ids_visited: A set of step IDs that have been visited to avoid cycles.

        Returns:
            None: This function modifies collected_paths in place by appending new paths
            as they are found.
        """
        # Case 1: If the steps_to_go is a custom_id string
        # This happens when a "next" of, for example, a IfFlowStepLink is targeting
        # a specific step by id
        if isinstance(steps_to_go, str):
            for i, step in enumerate(self.steps):
                # We don't need to check for 'id' as a link can only happen to a
                # custom id.
                if step.custom_id == steps_to_go:
                    self._go_over_steps(
                        self.steps[i:], current_path, completed_paths, step_ids_visited
                    )

        # Case 2: If steps_to_go is a list of steps
        else:
            for i, step in enumerate(steps_to_go):
                # 1. Check if the step is relevant for testable_paths extraction.
                # We only create new path nodes for ActionFlowStep and
                # CollectInformationFlowStep because these are externally visible
                # changes in the assistant's behaviour (trackable in the e2e tests).
                # For other flow steps, we only follow their links.
                # We decided to ignore calls to other flows in our coverage analysis.
                if not isinstance(step, (CollectInformationFlowStep, ActionFlowStep)):
                    self._handle_links(
                        step.next.links,
                        current_path,
                        completed_paths,
                        step_ids_visited,
                    )
                    continue

                # 2. Check if already visited this custom step id
                # in order to keep track of loops
                if step.custom_id is not None and step.custom_id in step_ids_visited:
                    if not completed_paths.is_path_part_of_list(current_path):
                        completed_paths.paths.append(copy.deepcopy(current_path))
                    return  # Stop traversing this path if we've revisited a step
                elif step.custom_id is not None:
                    step_ids_visited.add(step.custom_id)

                # 3. Append step info to the path
                current_path.nodes.append(
                    PathNode(
                        flow=current_path.flow,
                        step_id=step.id,
                        lines=step.metadata["line_numbers"],
                    )
                )

                # 4. Check if 'END' branch
                if (
                    len(step.next.links) == 1
                    and isinstance(step.next.links[0], StaticFlowStepLink)
                    and step.next.links[0].target == END_STEP
                ):
                    if not completed_paths.is_path_part_of_list(current_path):
                        completed_paths.paths.append(copy.deepcopy(current_path))
                    return
                else:
                    self._handle_links(
                        step.next.links,
                        current_path,
                        completed_paths,
                        step_ids_visited,
                    )

    def _handle_links(
        self,
        links: List[FlowStepLink],
        path: FlowPath,
        collected_paths: FlowPathsList,
        step_ids_visited: set,
    ) -> None:
        """Processes the next step in a flow.

        Potentially recursively calling itself to handle conditional paths and
        branching.

        Args:
            links: Links listed in the "next" attribute.
            path: The current path taken in the flow.
            collected_paths: A list of paths collected so far.
            step_ids_visited: A set of step IDs that have already been visited
                to avoid loops.

        Returns:
            None: Modifies collected_paths in place by appending new paths
            as they are completed.
        """
        steps = self.steps

        for link in links:
            # Direct step id reference
            if isinstance(link, StaticFlowStepLink):
                # Find this id in the flow steps and restart from there
                for i, step in enumerate(steps):
                    if step.id == link.target_step_id:
                        self._go_over_steps(
                            steps[i:],
                            copy.deepcopy(path),
                            collected_paths,
                            copy.deepcopy(step_ids_visited),
                        )

            # If conditions
            elif isinstance(link, (IfFlowStepLink, ElseFlowStepLink)):
                # Handling conditional paths
                target_steps: Union[str, List[FlowStep]]
                if isinstance(link.target_reference, FlowStepSequence):
                    target_steps = link.target_reference.child_steps
                else:
                    target_steps = link.target_reference

                self._go_over_steps(
                    target_steps,
                    copy.deepcopy(path),
                    collected_paths,
                    copy.deepcopy(step_ids_visited),
                )



from collections import defaultdict
from typing import Any, Dict, Text

from rasa.nlu.emulators.emulator import Emulator
from rasa.shared.nlu.constants import (
    ENTITIES,
    ENTITY_ATTRIBUTE_TYPE,
    ENTITY_ATTRIBUTE_ROLE,
    ENTITY_ATTRIBUTE_VALUE,
    ENTITY_ATTRIBUTE_END,
    ENTITY_ATTRIBUTE_START,
    TEXT,
    INTENT,
)


class WitEmulator(Emulator):
    """Emulates the response format of this wit.ai endpoint.

    More information about the endpoint:
    https://wit.ai/docs/http/20200513/#get__message_link
    """

    def normalise_response_json(self, data: Dict[Text, Any]) -> Dict[Text, Any]:
        """Transform response JSON to wit.ai format.

        Args:
            data: input JSON data as a dictionary.

        Returns:
            The transformed input data.
        """
        entities = defaultdict(list)
        for entity in data[ENTITIES]:
            entity_name = entity[ENTITY_ATTRIBUTE_TYPE]
            role = entity.get(ENTITY_ATTRIBUTE_ROLE, entity_name)
            entity_name_including_role = f"{entity[ENTITY_ATTRIBUTE_TYPE]}:{role}"
            normalized_entity: Dict[Text, Any] = {
                "confidence": entity.get("confidence_entity") or 1,
                "name": entity_name,
                "value": entity[ENTITY_ATTRIBUTE_VALUE],
                # Entity value before value was transformed (e.g. by synonym mapper)
                "body": data["text"][
                    entity.get(ENTITY_ATTRIBUTE_START, 0) : entity.get(
                        ENTITY_ATTRIBUTE_END, 0
                    )
                ],
                "start": entity[ENTITY_ATTRIBUTE_START],
                "end": entity[ENTITY_ATTRIBUTE_END],
                "role": role,
                "entities": [],
            }

            entities[entity_name_including_role].append(normalized_entity)

        return {"text": data[TEXT], "intents": [data[INTENT]], "entities": entities}
